import SEP.vector
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2004.12.2"
                                                                                


def dot_problem(dot1,dot2):
    """Give an error when doing dot product"""
    SEP.util.msg("Zero dot product result when dotting %s and %s"%(dot1,dot2))
    return None

class cgstep:
  """Conjugate-gradient step """

  def __init__(self,name):
    """Initialize a cgstep object

       name- ascii string to identify cgstep

    """
    self.name=name
    self.eps=.000001

  def alloc(self,mod,dat,restart):
    """allocate space (or load) previous step 

       mod     - a vector of the shape of the model
       dat     - a vector of the shape of the residual
       restart - whether or not we are restarting the inversion

       return 1 if successful, None if error 

    """
    self.s=mod.clone_space(name=self.name+"_s")
    self.ss=dat.clone_space(name=self.name+"_ss")
    if restart:
      self.s.load()
      self.ss.load()
    else:
      self.s.zero()
      self.ss.zero()
    self.forget="yes"

  def set_forget(self,a):
    """set the status of the forget variable 

        if a is true we do steepest descent
        if a is None do conjugate gradient step

    """ 
    self.forget=a

  def clean(self):
    """clean the vectors neeeded for the conjugate gradient """
    self.ss.clean()
    self.s.clean()

  def step(self,x,g,rr,gg):
    """Perform a conjugate gradient step


        x  - model vector
        g  - gradient vector
        gg - L g vector
        rr - residual vector

    """

    if self.forget:
      self.s.zero()
      self.ss.zero()
      beta=0.
      gdot=gg.dot(gg)
      if abs(gdot)< self.eps : return dot_problem("gg","gg")
      alfa=-gg.dot(rr)/gdot
      self.forget=None
    else:
      gdg=gg.dot(gg); #search plane by solving 2-by-2
      sds=self.ss.dot(self.ss);  # G . (R - G*alfa - S*beta) = 0
      gds=gg.dot(self.ss);   #S . (R - G*alfa - S*beta) = 0
      sdg=self.ss.dot(gg);   #S . (R - G*alfa - S*beta) = 0
      if gdg ==0.: return dot_problem("gg","gg")
      if sds==0.: return dot_problem("ss","gg")
      determ=abs(gdg*sds - gds*sdg)
      gdr=gg.dot(rr)
      sdr=self.ss.dot(rr)
      alfa= -( sds * gdr - gds * sdr ) / determ;
      beta =- (-gds * gdr + gdg * sdr ) / determ;

    self.s.scale_addscale(beta,g,alfa)       #update model step
    x.add(self.s)    #update solution
    self.ss.scale_addscale(beta,gg,alfa)     #update residual step
    rr.add(self.ss)  #update residual
    return 1

