import SEP.sepfile
import SEP.util
import SEP.spawn
import types
import string
import SEP.paths
import SEP.options
import SEP.parjob
import SEP.prog
import SEP.paths
import SEP.parfile
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.02.02"

def envelope_docs():
  par_list=[]
  par_list.append(SEP.options.parameter("tc1",0,doc="Half-width of smoother along axis 1"))
  par_list.append(SEP.options.parameter("tc2",0,doc="Half-width of smoother along axis 2"))
  par_list.append(SEP.options.parameter("freq",1,doc="Produces instanteous frequency instead of envelopes"))
  return par_list

class envelope(SEP.options.par_group,SEP.prog.prog):
  """Class to do envelope"""

  def __init__(self,name="interp"):
    """Initialize the Envelope program"""
    SEP.options.par_group.__init__(self,name)
    self.add_doc_params()

  def add_doc_params(self):
    """Parameters for Envelope"""
    self.add_doc_parameters(envelope_docs())
    self.add_doc_param("stdin",doc="Input SEP dataset")
    self.add_doc_param("verb","n",doc="Whether or not to be verbose")
    self.add_doc_param("stdout",doc="Output SEP dataset")

  def runtime_check_build(self):  
    """Check to make sure the parameters make sense"""

    S=SEP.sepfile.sep_file(self.param("stdin"))

    SEP.prog.prog.__init__(self,"%s/%s"%(SEP.paths.sepbindir,"Envelope"),
      "Envelope",self.return_args(),self.param("verb"))

class envelope_par(SEP.parjob.par_job):
  """Class to do Envelope"""

  def __init__(self,name="envelope"):
    """Initialize the Envelope program"""
    SEP.util.err("Parallel Envelope unimplemented")
    SEP.parjob.par_job.__init__(self,name)
    self.add_doc_params()

 
  def add_doc_params(self):
    """Parameters for Bandpass"""
    self.add_doc_params(envelope_docs())
    self.add_doc_param("nblock",doc="The number of blocks to break the data into ")
    self.add_doc_param("restart",doc="Whether or not we are restarting ",required=None)
    self.add_doc_param("intag",doc="Input SEP dataset (cmpx,cmpy,offset,offsety,z)[reversed for inv]")
    self.add_doc_param.parameter("outtag",doc="Output, SEP regular (z,angx,angy,cmpx,cmpy) [reversed for inv]")
    self.add_doc_param("dff_axis",doc="Axis to distribute the data along")
    self.prog_pars=["inverse","verb","maxmem","eps","push","nzpad"]
                                                                                                     

  def runtime_check_build(self):  
    """Check to make sure the parameters make sense"""

    restart=self.param("restart")
    S=SEP.sepfile.sep_file(self.param("intag"))
    if S.grid: SEP.util.err("Input should be regular SEP3D file")
    if S.headers: SEP.util.err("Input should be regular SEP3D file")

    self.inverse= SEP.par.logic(self.param("inverse"))
    
    #get the section parameters
    sect_pars,isect_dict,dff_in,dff_out=self.build_sect_params(S)

    par_files={}

    par_files["intag"]=SEP.parfile.distribute(name=self.param("intag"), dff_axis=dff_in,
     tag="<",usage="INPUT",isect_dict=isect_dict,restart=restart)

    par_files["outtag"]=SEP.parfile.distribute(name=self.param("outtag"),dff_axis=dff_out,
     tag=">",usage="OUTPUT",isect_dict=isect_dict,restart=restart)

    self.add_param_obj("files",par_files)
    order=[]
    for i in range(len(isect_dict.keys())): order.append(str(i))
    self.add_param_obj("order",order)
    self.add_param_obj("sect_pars",sect_pars)
    axes=self.param("axes")
    SEP.parjob.par_job.runtime_check_build(self)
    self.set_global_par_dict(self.return_args().return_basic_pars(self.prog_pars))


  def build_sect_params(self,sfile):
    """Build parameters for parallel job"""
    sect_pars={}
    isect_dict={}
    if self.inverse:
      if 1== int(sfile.axis(4)[0]) : 
        dff_in=4
        dff_out=2
      else: 
        dff_in=3
        dff_out=1
    else:
      if 1== int(sfile.axis(2)[0]): 
        dff_in=2
        dff_out=4
      else: 
        dff_in=1
        dff_out=3
    nblock=int(self.param("nblock"))
    for i in range(nblock):
      sect_pars[str(i)]=SEP.par.sep_pars(name=str(i))
      isect_dict[str(i)]=i
    return sect_pars,isect_dict,dff_in,dff_out


  def command_func(self,key,mlabel,par_args,job_args,tags):
    """Return the command to run given the job descriptor, job_string and file tags
                                                                                                     
       key      - the jobid
       mlabel   - the machine label we will be running on
       par_args - the parameters for the parallel jobs
       job_args - the parameters for job args
       tags     - the local file tags
                                                                                                     
       This should be overwritten when doing more complex jobs
                                                                                                     
    """

    off2ang="%s/OFF2ANG"%SEP.paths.sepbindir
    transp ="%s/Transp"%SEP.paths.sepbindir
    off_l=self.prog_list.prepare(off2ang,self.mach.mach_from_label(mlabel))
    tra_l=self.prog_list.prepare(transp ,self.mach.mach_from_label(mlabel))
    maxsize=self.param("maxmem")
    
    if tags[0][0]=="<": 
      input =tags[0]
      output=tags[1]
    else:
      input =tags[1]
      output=tags[0]

    if self.inverse:
      run="%s  %s sep.begin_par=1  %s %s| "%(off_l,input,
        string.join(par_args.return_sep_par_list()),
        string.join(job_args.return_sep_par_list()))
      run="%s %s  %s sep.end_par=1  maxsize=%s plane=13 reshape=2,4,5  %s"%(run,tra_l,output,maxsize,
        string.join(par_args.return_sep_par_list()))
    else:
      run="%s  %s sep.begin_par=1  maxsize=%s plane=13 reshape=2,4,5 %s | "%(tra_l,input,maxsize,
        string.join(par_args.return_sep_par_list()))
      run="%s %s  %s sep.end_par=1  %s %s "%(run,off_l,output,
        string.join(par_args.return_sep_par_list()),
        string.join(job_args.return_sep_par_list()))

    return run


