import SEP.sep_socket
import os,commands,sys,string,re
import SEP.par_log,pwd
import SEP.spawn
import SEP.paths
import SEP.util
import SEP.par_prog
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2004.12.2"
                                                                                

mpiN=0

class mpi:
  """A class to handle running mpi jobs"""
  def __init__(self,verb=None,startup=40,speed=12,tries=3,logdir=None):
      """Initialze the mpi object
 
          verb    - the verbosity level (defaults None)
          startup - the maximum time it takes to start a job on the node
          speed   - the minimum speed in transfreing (in megabytes per second)
          tries   - the number of tries before failing to distribute or collect
          logdir  - the directory to log MPI jobs

      """
      if SEP.paths.mpibindir:
        self.mpi_com=SEP.paths.mpibindir+os.sep+"mpirun"
      else:
        self.mpi_com="mpirun"
      if verb and int(verb) >5:self.verb=verb
      else: self.verb=None
      self.startup=startup
      self.speed=speed
      self.tries=tries
      self.log=None
      self.distrib={}
      if logdir: self.set_log(logdir)
      self.distrib_progs=SEP.par_prog.distrib_list()


  def clean(self):
    """Clean up from an  MPI jobs"""
    self.distrib_progs.clean()
       

  def send_executable(self,pars,nodes):
    """Make sure that the MPI executable is on each node

      pars  - the command line arguments
      nodes - the nodes we are running on
   
      return pars with the local executable name

    """
    prog=pars[0];
    self.distrib_progs.check_add_prog(prog)
    for node in nodes: pars[0]=self.distrib_progs.prepare(prog,node)
    return pars
    

  def run_job(self,pars,nodes,nbytes):
     """Run a MPI job

         pars  - the command line arguments
         nodes - the nodes to run
         nbytes- the number of bytes we are sending
     """
     pars=self.send_executable(pars,nodes)
     mach_file,com,stat_good=self.local_com(pars,nodes)
     itry=0
     err=1
     while itry< self.tries and err:
       logfile=self.logfile("MPI")
       err=SEP.spawn.run_wait(com,
         self.wait_time(nbytes,len(nodes)), logfile=logfile,error=None,verb=self.verb)
       sys.stdout.flush()
       if not err:
         sys.stdout.flush()
         if not os.path.isfile(stat_good): err=1
         elif not re.compile("junk").search( commands.getstatusoutput("grep junk %s"%stat_good)[1]):  err=1
       sys.stdout.flush()
       if err:
         SEP.util.msg("Trouble running %s "%com)
         try: f=open(logfile)
         except: SEP.util.err("Trouble opeining up log file=%s "%logfile)
         line=f.readline()
         while line:
           SEP.util.msg(line.rstrip())
           line=f.readline()
         f.close()
       a,b=commands.getstatusoutput("rm %s"%stat_good)
       itry=itry+1
     if err: return err
#     else: SEP.spawn.run_wait("rm %s"%mach_file);
     return None

  def local_com(self,pars,nodes):
    """Run the local command on pars on a series of nodes

        pars - the command line arguments
        nodes- the nodes to run on

    """
    mach=SEP.util.temp_file("mach")
    args=[self.mpi_com,"-np",str(int(len(nodes))+1),"-stdin /dev/null","-machinefile",mach]
    try: f = open (mach,"w")   #OPEN THE FILE
    except:
      SEP.util.err( "trouble opening temporary file "+mach)
    for host in nodes:
      f.write(host+"\n")
    f.close()
    args.extend(pars)
    stat_good=SEP.util.temp_file("goodMPI")
    args.append("stat_good=%s"%stat_good)
    args.append("verb=1")
    com=string.join(args)
    return mach,com,stat_good

  def wait_time(self,bytes,nmach):
    """The maximum amount it should take to run a given job

        bytes - the number of bytes we are transfering
        nmach - the number of machines we are running on

    """
    tm=self.startup*nmach #startup time from processes
    tm+=bytes/self.speed*2/1000000 #passing time 
    return tm

  def logfile(self,par):
     """Return the logfile we should send our output to """
     file="%s%s%s.%d"%(self.logdir,os.sep,par,SEP.mpi.mpiN)
     SEP.mpi.mpiN =SEP.mpi.mpiN+1
     return file
                                                                                  
  def set_log(self,log):
    """Set the logfile directory"""
    self.logdir=log
                                                                                  

