import SEP.util
import SEP.opt_vector
import time,os,string,types


def check_space(vec1,vec2):
    if type(vec1)!=type(vec2):
      SEP.util.msg("Vectors not of the same type")
      return 1
    if vec1.size()!=vec2.size():
      SEP.util.msg("Vectors not the same size")
      return 1
    return None


class operator:
  def __init__(self,name,domain,range,**kw):
    """Operator class"""
    self.domain=domain.clone_space()
    self.range=range.clone_space()
    self.kw=kw
    self.name=name
    self.verb=self.option("verb",0)
    self.nsteps=1
    self.msg=self.option("msg",name)

  def option(self,par,default=None):
    """Return an optional paramter if it exists, else None"""
    if not self.kw.has_key(par):  return default
    return self.kw[par]

  def adjoint(self,d,r,status,iter,add=None,restart=None):
    """Adjoint operation"""
    name="%d.%s_a"%(iter,self.name)
    if restart and name!=restart: return restart
    self.check_operator_logic(d,r)
    status.update_status(name,["started"])
    self.adjoint_op(d,r,add,restart)
    status.update_status(name,["finished"])
    if  restart: return None
    return restart

  def init_op(self,restart):
    """Initialize operator"""

  def adjoint_op(self,model,data,add,restart):
    """Function to run the adjoint operation"""
    SEP.util.err("Must define the adjoint operation")

  def forward_op(self,model,data,add,restart):
    """Function to run the forward operation"""
    SEP.util.err("Must define the forward operation")
    
  def job_desc(self,iter):
    """Return job description for an operator"""
    return ["%d.%s_f"%(iter,self.name)], ["%d.%s_a"%(iter,self.name)]
    
  def check_operator_logic(self,d,r):
    if check_space(self.domain,d):
       SEP.util.err("Domain vector %s not the write space for operator %s"%(d.name,self.name))
    if check_space(self.range,r):
       SEP.util.err("Range vector %s not the write space for operator %s"%(d.name,self.name))

  def forward(self,d,r,status,iter,add=None,restart=None):
    """Forward operation"""
    self.check_operator_logic(d,r)
    name="%d.%s_f"%(iter,self.name)
    if restart and name!=restart: return restart
    if self.verb ==2: print "%d iter forward operation %s"%(iter,self.msg)
    if self.verb >2: print "%d iter forward operation  %s model=%s data=%s"%(iter,self.msg,d.name,r.name)
    status.update_status(name,["started"])
    self.forward_op(d,r,add,restart)
    status.update_status(name,["finished"])
    if  restart: return None
    return restart

  def clean(self):
    """Nothing for default"""

  def nops(self):
    return self.nsteps 

class operator_run:
  """Class to create a simple out of core operator"""
  def __init__(self,name,**kw):
     self.prog=name
     if  self.option("pars"): self.pars=self.option("pars")
     else    : self.pars=SEP.par.sep_pars(name=name)
     self.kw=kw
     self.add_com=self.option("add_com","add=y")
     self.restart_com=self.option("restart_com","restart=y")
     self.adjoint_com=self.option("adjoint_com","adj=y")
     self.data_tag=self.option("data_tag","data=")
     self.model_tag=self.option("model_tag","model=")

  def option(self,par,default=None):
    """Return an optional paramter if it exists, else None"""
    if not self.kw.has_par(par):  return default
    return self.kw[par]

  def forward_pars(self,add=None,restart=None,extra=None):
    """Return the forward command"""
    par=SEP.par.sep_pars(name=self.prog)
    par.add_pars(self.pars)
    if add: par.add_string(self.add_com)
    if restart: par.add_string(self.restart_com)
    if extra  : par.add_string(extra.join())
    return par

  def par_base(self):
    """Return arguments as a list"""
    return self.pars.return_par_list()

  def adjoint_pars(self,add=None,restart=None,extra=None):
    """Return the adjoint command"""
    par=self.forward_pars(add,restart,extra)
    par.add_string(self.adjoint_com)
    return par

  def run_adjoint(self,model,data,add=None,restart=None,extra=None):
    par=self.adjoint_pars(add,restart,extra)
    par.add_string("%s%s  %s%s"%(self.model_tag,model,self.data_tag,data))
    com_out="%s %s"%(par.name,par.return_par_string())
    SEP.spawn.run_wait(com_out.split())

  def run_forward(self,model,data,add=None,restart=None,extra=None):
    par=self.forward_pars(add,restart,extra)
    par.add_string("%s%s  %s%s"%(self.model_tag,model,self.data_tag,data))
    com_out="%s %s"%(par.name,par.return_par_string())
    SEP.spawn.run_wait(com_out.split())

class oc_operator(operator,operator_run):
  """An operator class for simple operators"""
  def __init__(self,prog,domain,range,**kw):
    if kw.has_key("name"): name=prog
    else: name=os.path.basename(prog)
    operator.__init__(self,name,domain,range,**kw)
    operator_run.__init__(self,name,**kw)

  def forward_op(self,model,data,add,restart):
    extra=[]
    if model.extra_args: extra.extend(model.extra_args)
    if data.extra_args: extra.extend(data.extra_args)
    self.run_forward(model.name,data.name,add,restart,extra=extra)

  def adjoint_op(self,model,data,add,restart):
    extra=[]
    if model.extra_args: extra.extend(model.extra_args)
    if data.extra_args: extra.extend(data.extra_args)
    self.run_adjoint(model.name,data.name,add,restart,extra=extra)

class array_op(operator):
  """Complex array operator"""
  def __init__(self,name,array,nrow,ncol,restart=None,verb=0,msg=None):
    if len(array) != nrow*ncol: 
      SEP.util.err("Error: nrow=%d ncol=%d len(array)"%(nrow,ncol,len(array)))

    self.name=name
    self.op_list=array
    self.nrow=nrow
    self.ncol=ncol
    self.verb=verb
    if msg: self.msg=msg
    else: self.msg=name
    self.nop=len(array)

    self.domain_list=[]
    for irow in range(self.nrow):  self.domain_list.append(None)

    self.range_list=[]
    for icol in range(self.ncol):  self.range_list.append(None)

    i=0
    for icol in range(ncol):
      for irow in range(nrow):
        if self.op_list[i]:
          if self.range_list[icol]: 
            if check_space(self.range_list[icol],self.op_list[i].range):
              SEP.util.err("range vectors not equiv")
          else: self.range_list[icol]=self.op_list[i].range.clone_space()
          if self.domain_list[irow]: 
            if check_space(self.domain_list[irow],self.op_list[i].domain):
              SEP.util.err("domain vectors not equiv")
          else: self.domain_list[irow]=self.op_list[i].domain.clone_space()
        i=i+1

    for irow in range(nrow): 
      if not self.domain_list[irow]: SEP.util.err("Undefined domain vector %d"%irow)

    for icol in range(ncol): 
      if not self.range_list[icol]: SEP.util.err("Undefined range vector %d"%icol)
    
    if ncol >1: self.range=SEP.opt_vector.super_vector(self.name+".range",self.range_list)
    else: self.range=self.range_list[0].clone_space()

    if nrow >1: self.range=SEP.opt_vector.super_vector(self.name+".domain",self.domain_list)
    else: self.domain=self.domain_list[0].clone_space()

  def job_desc(self,iter):
    """Return job description for a matrix operator"""
    list_f=[]
    list_a=[]
    for i in range(0,self.nop,1):
      a,b=self.op_list[i].job_desc(iter)
      list_f.extend(a)
    for i in range(self.nop-1,-1,-1):
      a,b=self.op_list[i].job_desc(iter)
      list_a.extend(b)
    return list_f,list_a
    

  def forward(self,d,r,status,iter,add=None,restart=None):
    if self.verb >1: print "Forward array operation "+self.msg

    for icol in range(self.ncol):
      addit=add
      for irow in range(self.nrow):
        if self.ncol >1:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].forward(d.vecs[irow],r.vecs[icol],
             status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].forward(d,r.vecs[icol],
             status, iter,add=addit,restart=restart)
        else:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].forward(d.vecs[irow],r,
             status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].forward(d,r,
             status, iter,add=addit,restart=restart)
        addit=1
    return restart
    
  def adjoint(self,d,r,status,iter,add=None,restart=None):
    if self.verb >1: print "Adjoint array operation "+self.msg
    for irow in range(self.nrow-1,-1,-1):
      addit=add
      for icol in range(self.ncol-1,-1,-1):
        if self.ncol >1:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d.vecs[irow],r.vecs[icol],
              status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d,r.vecs[icol],
              status, iter,add=addit,restart=restart)
        else:
          if self.nrow >1:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d.vecs[irow],r,
              status, iter,add=addit,restart=restart)
          else:
            restart=self.op_list[self.nrow*icol+irow].adjoint(d,r,
              status, iter,add=addit,restart=restart)
        addit=1
    return restart
    
 
  def init_op(self,restart):
    """Initialize a collumn operator"""
    for op in self.op_list: op.init_op(restart)

  def clean(self):
    if self.verb >2: print "Clean matrix operation "+self.msg
    for op in  self.op_list: op.clean()


class chain_op(operator):
  def __init__(self,name,op_list,restart=None,verb=0, msg=None):
    self.name=name
    if not msg: self.msg=name
    else: self.msg=msg
    self.verb=verb

    self.op_list=op_list
    self.inter_vec=[]
    self.domain=self.op_list[0].domain.clone_space()
    self.range=self.op_list[len(self.op_list)-1].range.clone_space()
    self.nsteps=0
    self.nop=len(self.op_list)
    for op in op_list: self.nsteps+=op.nops()

    for i in range(1,len(self.op_list)):
      rlen=self.op_list[i-1].range.size()
      dlen=self.op_list[i].domain.size()
      if rlen!=dlen:
        SEP.util.err("domain and range vector mismatch for chain_op")
      self.inter_vec.append(self.op_list[i].domain.clone_space(name=name+str(i)))
      

  def job_desc(self,iter):
    """Return job description for a row operator operator"""
    list_f=[]
    list_a=[]
    for i in range(0,self.nop,1):
      a,b=self.op_list[i].job_desc(iter)
      list_f.extend(a)
    for i in range(self.nop-1,-1,-1):
      a,b=self.op_list[i].job_desc(iter)
      list_a.extend(b)
    return list_f,list_a
    

  def forward(self,d,r,status,iter,add=None,restart=None):
    if self.verb >1: print "Forward row operation "+self.msg
    restart=self.op_list[0].forward(d,self.inter_vec[0],status,iter,restart=restart)
    for i in range(0,len(self.inter_vec)-3):
      restart=self.op_list[i].forward(self.inter_vec[i],self.inter_vec[i+1],status,iter,restart=restart)
    restart=self.op_list[self.nop-1].forward(self.inter_vec[self.nop-2],r,status,iter,add=add,restart=restart)
    
  def adjoint(self,d,r,status,iter,add=None,restart=None):
    if self.verb >1: print "Adjoint row operation "+self.msg
    restart=self.op_list[self.nop-1].adjoint(self.inter_vec[len(self.inter_vec)-1],r,status,iter,restart=restart)
    for i in range(len(self.inter_vec)-3,0,-1):
      restart=self.op_list[i].adjoint(self.inter_vec[i-1],self.inter_vec[i],status,iter,restart=restart)
    restart=self.op_list[0].adjoint(d,self.inter_vec[0],status,iter,add,restart=restart)

  def clean(self):
    if self.verb >2: print "Clean row operation "+self.msg
    for op in  self.op_list: op.clean()
    for vec in self.inter_vec: 
      print vec.name
      vec.clean()

  def init_op(self,restart):
    """Initialize a collumn operator"""
    for op in self.op_list: op.init_op(restart)
    for vec in self.inter_vec: 
      if not restart: vec.zero()
      else: vec.load()


class scale_op(operator):
  def __init__(self,name,domain,range,scale,restart=None,verb=0):
     if domain.size() != range.size():
       SEP.util.err("domain and range must be the same size for scale_op")
     self.scale=scale
     operator.__init__(self,name,domain,range,verb=verb)
     
  def adjoint_op(self,d,r,add,restart=None):
    if add: sc=1.
    else: sc=0.
    d.scale_addscale(sc,r,self.scale)

  def forward_op(self,d,r,add,restart=None):
    if add: sc=1.
    else: sc=0.
    r.scale_addscale(sc,d,self.scale)



