import SEP.opt_vector
import SEP.util
import SEP.opt_operator
import SEP.status
import re,sys



class solver:

  def __init__(self,name,x,resid,op,stepper,niter,verb=0,restart=None):


    self.x=x
    self.rr=resid
    self.op=op
    self.niter=niter
    self.verb=verb
    self.iter0=0
    self.name=name
    self.stepper=stepper
    self.g=self.x.clone_space(name=self.name+"_g")
    self.gg=self.rr.clone_space(name=self.name+"_gg")
    if not restart:
      self.g.zero()
      self.gg.zero()
    else:
      self.g.load()
      self.gg.load()
    self.vect_list.extend([self.g,self.gg])
    self.stepper.alloc(x,resid,restart)
    job_list=self.calc_work()
    self.status=SEP.status.sep_status(self.name+".stat",job_list)
    self.restart=restart
    self.op.init_op(restart)
    self.restart_it()

  def clean(self):
    self.stepper.clean()
    for vec in self.vect_list: 
      vec.clean()
    self.op.clean()

  def restart_it(self):
    #update status
    restart_list=self.status.restart(self.restart,["finished"],["started","restart"])
    if restart_list:
      iterS=re.compile("^(\d+)\.").search(restart_list[0])
      if  not iterS: SEP.util.err("Unknown status parameter %s"%restart_list[0])
      iter=int(iterS.group(1))
      self.iter0=iter+1
      if re.compile("_a$").search(restart_list[0]): step=0
      elif re.compile("_f$").search(restart_list[0]): step=1
      elif re.compile("step$").search(restart_list[0]): step=2
      else: SEP.util.err("unknown parameter %s"%restart_list[0]);
      if iter >0: self.stepper.set_forget(None)
      if step==0: 
        self.op.adjoint(self.g,self.rr,self.status,iter,restart=restart_list[0])
        self.op.forward(self.g,self.gg,self.status,iter)
      if step==1: 
        self.op.forward(self.g,self.gg,self.status,iter,restart=restart_list[0])
      self.status.update_status(str(iter)+".step",["started"])
      if not self.stepper.step(self.x,self.g,self.rr,self.gg): return None
      self.status.update_status(self.name+".step",["finished"])
      

  def calc_work(self):
    job_list=[]
    for iter in range(self.niter):
      f,a=self.op.job_desc(iter)
      job_list.extend(a)
      job_list.extend(f)
      job_list.append(str(iter)+".step")
    return job_list
      

  def step(self,iter):
    self.op.adjoint(self.g,self.rr,self.status,iter)
    self.op.forward(self.g,self.gg,self.status,iter)
    self.status.update_status(str(iter)+".step",["started"])
    if not self.stepper.step(self.x,self.g,self.rr,self.gg): return None
    self.status.update_status(str(iter)+".step",["finished"])
    return 1

  def report(self,iter):
    dd=self.rr.dot(self.rr)
    dx=self.x.dot(self.x)
    dg=self.g.dot(self.g)
    if iter==0: self.scale=dd/1000. 
    print "iter=%d r(dat)=%f,%f r(x)=%f,%f r(grad)=%f,%f"%(iter,dd.real,dd.imag,dx.real,dx.imag,dg.real,dg.imag)

  def iter(self):
    self.status.lock_it(read=None)
    for iter in range(self.iter0,self.niter):
      if not self.step(iter): return None
      if self.verb>0: self.report(iter) 
    self.status.unlock_it()
    return 1

class solver_smp(solver):

  def __init__(self,name,m,d,op,stepper,niter,wop=None,
       m0=None,resd=None,verb=None,restart=None):

    if wop:
      self.rr=wop.range.clone_space(name=name+".smp_rr")
      if restart: self.rr.load()
      else:
        self.rr.zero()
        wop.forward(d,rr)  #rr= W d
        self.rr.scale(-1.)      #rr= - Wd
      self.chain_op=SEP.opt_operator.chain_op(name+".WL",[op,wop],restart)
    else:
      self.rr=d.clone_space(name=name+".smp_rr") # rr=d
      if restart: self.rr.load()
      else:
        self.rr.zero()
        self.rr.scale_add(-1.,d)   # rr=-d 
        self.chain_op=op
    self.vect_list=[self.rr]

    if m0 and not restart:
      m.zero()
      m.add(m0)
      self.chain_op(m,self.rr,add="yes")

    self.m=m
    solver.__init__(self,name,self.m,self.rr,self.chain_op,stepper,niter,verb,restart)

      

  
  def solve():
    if not self.iter(): return None
    return 1

class solver_reg(solver):
  def __init__(self,name,m,d,op,reg,stepper,niter,eps,wop=None,rm0=None,
       m0=None,resd=None,verb=None,restart=None):

    self.verb=verb
    self.status=SEP.status.sep_status(name+".stat",name)

    if wop:
      self.rd=wop.range.clone_space(name=name+"_rd")
      if restart: self.rd.load()
      else: 
        self.rd.zero()
        wop.forward(d,rr)       #rd= W d
        self.rd.scale(-1.)      #rd= - Wd
      self.chain_op=SEP.opt_operator.chain_op(name+".WO",[op,wop])
    else:
      self.rd=d.clone_space(name=name+"_rd") # rd=d
      if restart: self.rd.load()
      else: 
        self.rd.zero()  
        self.rd.scale_add(-1.,d) #rd=-d
      self.chain_op=op

    self.rm=reg.range.clone_space(name=name+"_rm")
    if  restart: self.rm.load()
    else:
      self.rm.zero()
      if rm0: self.rm.add(rm0)
    self.vect_list=[self.rm,self.rd]

    self.scale_op=SEP.opt_operator.scale_op("eps_reg",m,m,eps,verb=self.verb)
    self.reg_chain=SEP.opt_operator.chain_op("eA",[reg,self.scale_op])
    self.m=m

    if not restart: self.m.zero()
    if m0 and not restart:
      self.m.add(m0)    # m=m0
      self.chain_op(self.m,self.rd,add=1)
      self.reg_chain(self.m,self.rm,add=1)
      
    self.rr=SEP.opt_vector.super_vector(name+".rr",[self.rd,self.rm])
    self.op=SEP.opt_operator.array_op(name+".main",[self.chain_op,self.reg_chain],1,2)
    solver.__init__(self,name,self.m,self.rr,self.op,stepper,niter,verb,
     restart)

  def solve(self):
    if not self.iter(): return None
    return 1

class solver_prec(solver):
  def __init__(self,name,m,d,op,prec,stepper,niter,eps,wop=None,rm0=None,
       p0=None,resd=None,verb=None,restart=None):

    self.verb=verb

    if wop:
      self.rd=wop.range.clone_space(name=name+"_rd")
      if restart: self.rd.load()
      else: 
        self.rd.zero()
        wop.forward(d,rr)       #rd= W d
        self.rd.scale(-1.)      #rd= - Wd
      self.chain_op=SEP.opt_operator.chain_op(name+".WLP",[prec,op,wop])
    else:
      self.rd=d.clone_space(name=name+"_rd")
      if restart: self.rd.load()
      else: 
        self.rd.zero()
        self.rd.scale_add(-1.,d)   # rd=-d 
      self.chain_op=SEP.opt_operator.chain_op(name+".LP",[prec,op])

    self.rm=prec.domain.clone_space(name=name+"_rm")
    if restart: self.rm.load()
    else:
      self.rm.zero()
      if rm0: self.rm=rm0.add(rm0)

    self.prec=prec
    self.m=m
    if not restart: self.m.zero()

    self.p=prec.domain.clone_space(name=name+"_p")
    if restart: self.p.load()
    else:
      self.p.zero()
      if p0: self.p.add(p0)
      
    self.scale_op=SEP.opt_operator.scale_op(name+".eps",self.p,self.m,eps,restart=restart,verb=self.verb)
      

    self.vect_list=[self.rd,self.rm,self.p]
    self.rr=SEP.opt_vector.super_vector(name+".rr",[self.rd,self.rm])
    self.op=SEP.opt_operator.array_op(name+".main",[self.chain_op,self.scale_op],1,2)
    solver.__init__(self,name,self.p,self.rr,self.op,stepper,niter,verb,restart)

  def calc_work(self):
     job_list=solver.calc_work(self)
     job_list.append("-1."+self.prec.name+"_f")
     return job_list

  def solve(self):
    if not self.iter(): return None
    self.prec.forward(self.p,self.m,self.status,-1)
    return 1
