import os
import sys
import string
import SEP.par
import SEP.util
import SEP.spawn
import types
import SEP.status

__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.1.25"
                                                                                


class parameter:
  """Class to hold a parameter description"""
  def __init__(self,param,default=None,doc=None,required=1):
     """Initialize a parameter

        param       - the parameter descriptor
        default     - the default value for the parameter
        doc         - a description for the parameter

     """
     self.param=param
     self.default=default
     self.doc=doc
     self.required=required
     if type(self.default) is not types.NoneType: self.required=None
     

  def clone(self):
    """Clone the parameter"""
    return parameter(self.param,self.default,self.doc,self.required)

  def return_doc(self,prefix=""): 
     """Return a string describing the parameter"""
     if type(self.default) is not types.NoneType:
       if self.doc: 
         st="%s%s - [%s]  %s "%(prefix,self.param,
             str(self.default),self.doc)
       else:
         st="%s%s - [%s] "%(prefix,self.param,str(self.default))
     else:
       if self.doc: 
         st="%s%s - %s "%(prefix,self.param,self.doc)
       else:
         st="%s%s "%(prefix,self.param)
     if not self.required: st="%s"%st
     wds=string.split(st)
     lst=["      "]
     i=0
     for wd in wds:
       if len(lst[i]) + len(wd) >70:
         lst.append("       ")
         i=i+1
       lst[i]=lst[i]+"%s "%wd
     return lst

  def set_default(self,d):
    """Set the default value for this parameter"""
    self.default=d

  def return_default(self):
    """Return the default value"""
    return self.default

  def  runtime_check_build(self):
    """Do any runtime checking for the flow"""


  def return_required(self):
    """Return whether the parameter is required"""
    return self.required

class par_group:
  """A group of parameters """

  def __init__(self,name,is_job=1):
     """Initialize a class for parameters """
     self.my_name=name
     self.par_list={}
     self.prog_opt={}
     self.local_dict={}
     self.is_job=is_job
     self.class_args=None
                                                                                
  def del_doc_param(self,par):
     """Delete a parameter from the self-doc"""
     if self.par_list.has_key(par): del self.par_list[par]
                                                                                
  def add_doc_param(self,par,default=None,doc=None,required=1):
    """Add a parameter to the self doc
                                                                                
        par         - the parameter descriptor
        default     - the default value for the parameter
        doc         - a description for the parameter
                                                                                
    """
    self.par_list[par]=parameter(par,default,doc,required)

  def add_doc_parameters(self,par_group):
     """Add a series of parameters to the documentation"""
     for p in par_group:   self.par_list[p.param]=p
     

  def clean_files(self):
    """Clean files"""

  def read_params(self,args,prefix="",delete=None):
      """Read the parameters for the par object

         args -  the program arguments to check against (SEP.par.sep_pars)
         prefix- the prefix associated with this parameter class
         delete- whether or not to delete the parameters from args that we need

         return err,args

           err        - whether (1) or not (None) all arguments were set
           args       - the command line args, possibly without arguments from this group
      
      """
      error=None
      self.class_args=SEP.par.sep_pars(name=self.my_name)
      for par,arg in self.par_list.items():
         d=arg.return_default()
         if type(d)!=types.NoneType: v,e=args.par("%s%s"%(prefix,par),d)
         else: v,e=args.par("%s%s"%(prefix,par))
         if e==0 and arg.return_required(): 
           error=1
           lines=arg.return_doc(prefix)
           for line in lines: SEP.util.msg(line)
         else:
           self.class_args.add_param(par,v)
         if delete: args.del_par("%s%s"%(prefix,par))
      return error,args


  def build_check_params(self):
    """Build program parameters based on other program parameters"""
      
  def runtime_check_build(self):
    """Check to make sure everything is setup before running"""

   
  def set_default(self,par,default):
     """Set the default for a parameter in this parameter group"""
     if not self.par_list.has_key(par):
        SEP.util.err("Attempt to set default %s which doesn't exist "%par)
     self.par_list[par].set_default(default)

  def param(self,par,error=1):
     """Retrieve a parameter from the local parameter list"""
     if not self.class_args:
       SEP.util.err("Arguments have not been set (read_params not called):%s"%self.my_name)
     if self.local_dict.has_key(par): return self.local_dict[par]
     return self.class_args.param(par,error=error)
       

  def add_string(self,parstring):
    """Add a parameter to the par_block"""
    if not self.class_args:
      SEP.util.err("Arguments have not been set (read_params not called):%s"%self.my_name)
    self.class_args.add_string(parstring)

  def del_par(self,par):
    """Delete a parameter to the par_block"""
    if not self.class_args:
      SEP.util.err("Arguments have not been set (read_params not called):%s"%self.my_name)
    self.class_args.del_par(par)

  def add_param_obj(self,par,val):
    """Add a parameter that is an object"""
    self.local_dict[par]=val

  def logic(self,par):
    """Add a parameter to the par_block"""
    if not self.class_args:
      SEP.util.err("Arguments have not been set (read_params not called):%s"%self.my_name)
    a=self.class_args.logic(par)
    return self.class_args.logic(par)
     
  def add_param(self,par,val):
    """Add a parameter to the par_block"""
    if not self.class_args:
      SEP.util.err("Arguments have not been set (read_params not called):%s"%self.my_name)
    self.class_args.add_param(par,val)
     
  def return_args(self):
    """Return the arguments for this parameter block"""
    if not self.class_args:
      SEP.util.err("Arguments have not been set (read_params not called):%s"%self.my_name)
    return self.class_args


  def return_name(self):
    """Return the name for this parameter block"""
    return self.my_name
                                                                                
  def doc(self,prefix=""):
     """Return the documentation for this section"""
     require=[]
     defaulted=[]
     option=[]
                                                                                
     for par,struct in self.par_list.items():
        if type(struct.default) is not types.NoneType : defaulted.append(par)
        elif struct.required: require.append(par)
        else:  option.append(par)
     doc=[]
     if require: doc.extend(["     Required"])
     for par in require: doc.extend(self.par_list[par].return_doc(prefix))
     if option: doc.extend(["     Optional"])
     for par in option : doc.extend(self.par_list[par].return_doc(prefix))
     if defaulted: doc.extend(["     Defaulted"])
     for par in defaulted : doc.extend(self.par_list[par].return_doc(prefix))
     return doc

  def set_prog_options(self,prog):
    """Set the options for each component of the program"""
    self.prog_opt=prog

  def return_prog_opt(self,sect):
     if not self.prog_opt.has_key(sect): return None
     return self.prog_opt[sect]
   


class flow:
  """A class for a combination of different modules [e.g. programs]"""
  def __init__(self,name,par_blocks=[],prefixes={},flows=[],restartable=None,order=[]):
    """A Class for a series of modules"""
    self.my_name=name
    self.par_blocks= par_blocks
    self.flows=flows
    self.prefixes=prefixes
    self.order=order
    self.restartable=restartable
    self.is_job=1

  def get_options(self,par_dict,args,prefix="",delete=None):
    """Return the options for this flow"""
    order,op_dict,type_dict=self.return_job_order()
    for o in order:
      if not type_dict.has_key(o):
        SEP.util.err("Structure %s doesn't exist"%o)
      if type_dict[o][0]=="F":
        if self.prefixes.has_key(op_dict[o].my_name): pre=self.prefixes[op_dict[o].my_name]
        else: pre=""
        par_dict,args=op_dict[o].get_options(par_dict,args,"%s%s"%(prefix,pre),delete)
      else:
        name=op_dict[o].return_name()
        if self.prefixes.has_key(name): pfx="%s%s"%(prefix,self.prefixes[name])
        else: pfx=prefix
        err,args=op_dict[o].read_params(args,pfx,delete)
        if err: SEP.util.err("Parameter problem in %s"%name)
        op_dict[o].set_prog_options(par_dict)
        op_dict[o].build_check_params()
        par_dict[name]=op_dict[o]
    return  par_dict,args

  def return_name(self):
    """Return the name for this parameter block"""
    return self.my_name

  def check_build_restartable(self,restart):
    """Build the list of commands to run"""
    restart_list=[]
    if self.restartable:
      job_list=self.return_job_list()
      self.status=SEP.status.sep_status(self.my_name+".stat",job_list)
      restart_list=self.status.restart(restart,
           ["finished"],["started","restart"])
      return restart_list
    else: return None

  def return_job_order(self):
    """Return the order in which jobs should be run"""
    if not self.order:
      order=[]
      for f in self.flows: order.append(f.my_name)
      for p in self.par_blocks: order.append(p.my_name)
    else: order=self.order
    op_dict={}; type_dict={} #could be done with type function
    for f in self.flows: 
      op_dict[f.my_name]=f
      type_dict[f.my_name]="FLOW"
    for p in self.par_blocks: 
      op_dict[p.my_name]=p
      type_dict[p.my_name]="PAR_GROUP"
    return order,op_dict,type_dict

  def get_name(self):
    """The name of the job to run"""
    return self.my_name
  
  def return_job_list(self):
    """Return the list of jobs for this job"""
    job_list=[]
    for f in self.flows: job_list.append(f.get_name())
    for block in self.par_blocks: job_list.append(block.return_name())
    return job_list
    
  def add_options(self,p,par_dict,args,delete):
    """Add options for the given flow"""
    for block in self.par_blocks:
       name=block.return_name()
       if self.prefixes.has_key(name): pre=self.prefixes[name]
       else: pre=""
       prefix="%s%s"%(p,pre)  
       err,args=block.read_params(args,prefix,delete)
       if err: SEP.util.err("Parameter problem in %s"%block.my_name)
       block.set_prog_options(par_dict)
       block.build_check_params()
       par_dict[name]=block
    return  par_dict,args
  def add_par_block(self,par):
    """Add a parameter block to the program"""
    self.par_blocks.append(par)
  def add_prefix(self,prefix,group):
     """Set a prefix for a parameter group"""
     self.prefixes[group]=prefix

  def  runtime_check_build(self):
    """Runtime check build"""
    order,op_dict,type_dict=self.return_job_order()
    for o in order:
      if  not op_dict.has_key(o):
        SEP.util.err("Trouble finding %s doesn't exist in flows or par groups"%o)
      op_dict[o].runtime_check_build() 

  def  run_job(self,restart=None):
    """Run the job(s) comprising the program"""
    restartit=self.check_build_restartable(restart)
    order,op_dict,type_dict=self.return_job_order()
    for o in order:
      if  not op_dict.has_key(o):
        SEP.util.err("Trouble finding %s doesn't exist in flows or par groups"%o)
      doit=1
      restart=None
      if self.restartable: 
         if restartit.count(o)==1: 
           op_dict[0].add_param("restart")
           restart=1
         if  self.status.get_key_status_elem(o,0)[0:7]=="finished": doit=0
      if doit:
        if type_dict[o][0] != "F": op_dict[o].runtime_check_build() 
        if self.restartable: self.status.update_status(o,["started"],lock=1)
        if op_dict[o].is_job:op_dict[o].run_job(restart) 
        op_dict[o].clean_files()
        if self.restartable: self.status.update_status(o,["finished"],lock=1)

  def clean_files(self):
    """Clean any files from the job"""
    if os.path.isfile("rm %s.stat"%self.my_name):
      SEP.spawn.run_wait("rm %s.stat"%self.my_name)


  def check_build_order(self):
    """Build the order in which jobs should be run"""


  def return_args(self,name=None): 
     """Return the arguments for a given par group"""
     if not name: return self.args
     if not self.par_dict.has_key(name):
       SEP.util.err("Request for undefined parameter group %s"%name)
     return self.par_dict[name].return_args()

  def doc_pars(self,prefix=""):
    """Return the docs for a given parameter class"""
    lines=[]
    order,op_dict,type_dict=self.return_job_order()
    for o in order:
      if not type_dict.has_key(o):
        SEP.util.err("Structure %s doesn't exist"%o)
      if type_dict[o][0]=="F":
        if self.prefixes.has_key(op_dict[o].my_name): pre=self.prefixes[op_dict[o].my_name]
        else: pre=""
        lines.extend(op_dict[o].doc_pars("%s%s"%(prefix,pre)))
      else:
        name=op_dict[o].return_name()
        if self.prefixes.has_key(name): pfx="%s%s"%(prefix,self.prefixes[name])
        else: pfx=prefix
        b=op_dict[o].doc(pfx)
        if b:
          lines.extend(["  %s"%op_dict[o].return_name()])
          lines.extend(b)
    return lines
    

class prog(flow):
  """A class for documenting the options in a program"""


  def __init__(self,summary,usage,par_blocks=[],description=[],comments=[],examples=[],
     args=None,name=None,flows=[],prefixes={},restartable=None,order=[]):
    """Initialize a program object

       summary    - A one line summary for the program
       usage      - Example usage
       par_blocks - A list of par_groups that make up the arguments
       description- A list describing the program
       comments   - A list describing comments for the program
       examples   - A list with examples for the
       args       - The arguments for the program  (defaults to sys.argv)
       name       - The name of the program   (defaults to sys.argv[0])
       prefixes   - A dictionary with  par_block.name:prefix defaults to "" for all
       restartable- Whether or not to make the program restartable
       order      - The order to run things. Defaults to flows then par_blocks

    """
       
    self.summary   =summary
    self.usage     = usage
    self.description=description
    self.comments   =comments
    self.examples   =examples
    self.par_dict={}
    if args: 
      self.args=args
    else:
      self.args=SEP.par.sep_args()

    if self.args.param("print_all",delete=1,error=None): SEP.spawn.print_all=1
    if self.args.param("debug_py",delete=1,error=None): SEP.util.debug=1
    self.restart=self.args.param("restart",error=None)

    if not name: name=os.path.basename(sys.argv[0])
    flow.__init__(self,name=name,par_blocks=par_blocks,prefixes=prefixes,flows=flows,restartable=restartable,order=order)
    self.prefixes=prefixes

    if len(sys.argv) ==1:
      d=self.doc()
      for line in d: print "%s "%line
      sys.exit(-1)



  def add_description(self,des):
    """Add a line to to the description for the program"""
    self.description.append(des)
  def add_comment(self,com):
    """Add a comment line to the program"""
    self.comments.append(com)
  def add_example(self,examp):
    """Add a line to the examples for the program"""
    self.examples.append(examp)

  def get_options(self,delete=None):
     """Read/build the parameters for the program"""
     self.par_dict={}
     self.par_dict,self.args=flow.get_options(self,self.par_dict,self.args,
       "",delete)


    
     
     
  def return_prog_args(self):
    """Return the arguments for the all par groups"""
    total_args=SEP.par.sep_pars(name=self.my_name)
    for block in self.par_blocks:
       total_args.add_pars(block.return_args())
    return total_args

  def doc(self):
    """Return the documentation for the program"""
    lines=["%s -  %s"%(self.my_name,self.summary),""]
    lines.extend(["USAGE","  %s"%self.usage])
    lines.extend(["PARAMETERS"])

    lines.extend(self.doc_pars(""))
    if self.description:
      lines.extend(["DESCRIPTION",""])
      for line in self.description: lines.append(line)
    
    if self.examples:
      lines.extend(["EXAMPLES",""])
      for line in self.examples: lines.append(line)
    
    if self.comments:
      lines.extend(["COMMENTS",""])
      for line in self.comments: lines.append(line)
    return lines

