import SEP.sep_socket
import SEP.superset
import SEP.datapath
import SEP.mpi
import os
import sys
import commands
import SEP.util
 
class parfile:
  """ Define a sep parallel file"""
  def __init__(self,
     name,  #file name
     tag,  #tag name
     isect_dict, #dictionary section ident:isect
     usage,  #usage INPUT,OUTPUT
     mach,  #machine functions
     jobs,    #Job running class
     axis=None,  #axis
     type="BLOCK", #distribution type BLOCK,SPREAD,COPY
     combine="YES",#combine the dataset when finished
     restart=None,  #whether or not we are restarting a job
     shared="NO",  #whether or not the COPY distributed file are shared on a node
     verb=None,    #whether or not to be verbose
     reuse_par=None,    # parameter to add if the file is being reused
     status=None,     #status. Only necessary if restarting
     remove=None     #remove parallel file upon completion (yes if combine="YES")
     ):
 
     self.not_first=None
     self.set_init_pars(name,tag,axis,type,usage,combine,restart,shared,isect_dict,verb,reuse_par,jobs,status,mach,remove)
     dead=0
     self.check_param_logic(dead)
     self.check_file(dead)
     if dead ==1: SEP.util.err(" ")
     if self.verb>2: SEP.util.msg("    Finished initializing %s"%self.name)




  def tags(self,mach_list):
    """Get local section tag given machine  list section list"""
    sect_names,new_list=self.get_section_names(mach_list)
    tag_list={}
    for key in mach_list.keys():
      self.sect_dict[key]=sect_names[key]
      self.mach_dict[key]=mach_list[key]
      tag_list[key]=self.tag+sect_names[key]
      if self.reuse_par != None:
        if not new_list.has_key(key): 
           a=tag_list[key]
           tag_list[key]+=" "+self.reuse_par
    return tag_list

  def finish(self):
    """Finish processing a file"""
    nout=0
    if self.usage == "INPUT" : return
    for n in self.naxis: nout=int(n)+int(nout)
    n,o,d,label,unit=self.par_file.axis(self.axis)
    self.par_file.set_axis(self.axis,nout,self.oaxis,d,label,unit)
    if self.type!="COPY": 
      self.file.set_axis(self.axis,nout,self.oaxis,d,label,unit)
    else: 
      self.file.set_axis(self.axis,1,0.,1.,"none","none")
      self.par_file.set_axis(self.axis,1,0.,1.,"none","none")
    self.par_file.compress_dff()
    self.par_file.write_file()
    if self.combine == "YES": self.file.write_file()

  def remove_it(self):
    if self.remove == "NO": return
    dict={}
    dict["program"]=SEP.paths.sepbindir+"/Srm3d"
    if self.remove=="YES": self.par_file.remove(dict)
    if self.verb>3: dict["verb"]=1
     

  def done(self,key_list):
    """Mark that a section has been processed"""

    if self.usage == "INPUT": return
    for key in  key_list: self.mark_done(key)
    self.par_file.write_file()
    if self.combine == "YES": self.file.write_file()

  def dead(self,mach):
    """Mark that a node has died"""
    for key,val in self.mach_dict:
      if val == mach:
        isect=self.isect_dict[key]
        self.naxis[isect]=0
        self.par_file.set_fake(isect,1)

  def mark_done(self,key):
    """Mark section finished and update files"""
    sect=self.sect_dict[key]
    mach=self.mach_dict[key]
    isect=self.isect_dict[key]
    temp_file=SEP.sepfile.sep_file(remote=mach+":"+sect)
    if not self.par_file.isect_by_name(sect):
      self.par_file.sections[key]=SEP.superset.sep_section(
        isect,mach,sepfile=temp_file,name=sect)
      self.par_file.add_dff_par("mach_sect"+str(isect),mach)
      self.par_file.add_dff_par("tag_sect"+str(isect),sect)
      self.par_file.set_fake(isect,0)
    ndim=temp_file.ndims()
    n,o,d,label,unit=temp_file.axis(self.axis)
    self.naxis[isect]=n
    if isect == 0: self.oaxis=o
    for i in range(ndim):
      iax=i+1
      n,o,d,label,unit=temp_file.axis(iax)
      if iax == self.axis and self.type != "COPY":
        if self.type == "BLOCK":
          o=float(o)-float(d)*int(isect)*int(n)
          n=int(n)*int(self.nsect)
        if self.type == "SPREAD":
          o=float(o)-float(d)*int(isect)
          d=float(d)/float(self.nsect)
          n=int(n)*int(self.nsect)
      self.par_file.set_axis(iax,n,o,d,label,unit)
    if self.combine == "YES": self.file.set_axis(iax,n,o,d,label,unit)
    self.par_file.write_file()
     
  def get_section_names(self,mach_list):
    """Get local section names given machine list section list"""
    mach_dist_list={}
    sect_dist_list={}
    isect_dist_list={}
    sect_list={}
    new_list={}
    for key,mach in mach_list.items():
      sect, dist,new=self.get_name(self.isect_dict[key],mach)
      if dist:  #if we need to distribute it
        mach_dist_list[key]=mach_list[key]
        sect_dist_list[key]=sect
        isect_dist_list[key]=self.isect_dict[key]
      if new: new_list[key]=sect
      sect_list[key]=sect
      self.sect_dict[key]=sect
      self.mach_dict[key]=mach_list[key]
    if len(mach_dist_list) >0:
#      imap=self.par_file.compress_dff()
      imap=isect_dist_list
      ilist={}
      i=0
      for key,val in imap.items():
         ilist[key]=val
         i+=1
      if self.verb>1: 
        SEP.util.msg("    Distributing  tag=%s %s"%(self.name,mach_dist_list))
      self.par_file.write_file()
      self.mpi.distribute(self.par_file.name,self.file.name,
        ilist,mach_dist_list, self.axis,self.transfer_size(isect_dist_list))
      for i in imap.keys(): self.par_file.set_fake(i,0)
      self.par_file.write_file()
    return sect_list,new_list

  def get_name(self,isect,mach):
    """Get the section name given the machine and section number"""
    new=None
    create=None
    if self.type=="COPY" and  self.shared=="YES":
      ilist,list=self.par_file.isect_tag_by_host(mach)
      if ilist == None: new=1
      else: sect_out=list[0]
    else: 
      mach_out,sect_out=self.par_file.mach_tag_by_isect(isect)
      if mach_out==None: new=1
      elif mach!=mach_out:  new=1
    if new:
      sect_out=self.par_file.create_section(mach,isect,1)
      if  self.usage=="INPUT": create=1
    return  sect_out,create,new
 
      
  def collect(self,key_list,extra_args):
    """Combine the sections in isect_list to output image"""
    isect_list={}
    mach_list=[]
    if self.usage == "INPUT": return
    for key in key_list: isect_list[key]=self.isect_dict[key]
    for key in key_list: mach_list.append(self.mach_dict[key])
    if self.verb>0: 
      SEP.util.msg("Tag=%s combining sections %s "%(self.name,key_list))
    if self.combine=="YES":
      self.mpi.collect(self.par_file.name,self.file.name,
        isect_list.values(),mach_list,self.transfer_size(isect_list),extra_args)

  def transfer_size(self,isect_list):
    """return the number of bytes that are transfered"""
    nsect=len(isect_list.keys())
    n=8*nsect  #assume complex
    for i in range(self.file.ndims()):
      if i+1 !=self.par_file.dff.axis or self.type!="COPY" :
        n=n*int(self.file.axis(i+1)[0])
    if self.type!="COPY":
      if self.usage == "INPUT": 
        n=n*int(self.par_file.sect_axis(int(isect_list.values()[0]))[0])
      else:
        n=n*int(self.par_file.sections.values()[0].axis(self.par_file.dff.axis)[0])
    return n

##########################
#INITIALIZATION FUNCTIONS#
##########################
                                                                                       
  def set_init_pars(self,name,tag,axis,type,usage,combine,restart,shared,isect_dict,verb,reuse_par,jobs,status,mach,remove):
     """ Initialize structure parameters"""
     self.file=None     #non-parallel sepfile
     self.name=name
     self.tag=tag
     self.reuse_par=reuse_par
     self.verb=verb
     self.axis=axis
     self.type=type
     self.usage=usage
     self.combine=combine
     self.isect_dict=isect_dict
     self.mpi=SEP.mpi.mpi(jobs,verb=verb)
     self.shared=shared
     self.restart=restart
     self.sect_dict={}
     self.mach_dict={}
     self.naxis=[]
     self.oaxis=0.
     self.jobs=jobs
     self.status=status
     self.mach=mach
     if remove: self.remove=remove
     elif self.combine=="YES": self.remove="YES"
     else: self.remove="NO"
                                                                                       
  def check_param_logic(self,dead):
     """Check to make sure required parameters are set"""
     self.nsect=len(self.isect_dict)
     if self.tag == None: SEP.util.err( "Must supply tag")
     if self.restart and not self.status: 
       SEP.util.err("Must provide status if restarting")
     if self.axis == None:
       if self.type!="COPY": SEP.util.err( "Must supply axis")
       else: self.axis=9
     if self.type != "BLOCK" and self.type!="SPREAD" and self.type!="COPY":
       SEP.util.err( "Distribution type must be BLOCK,SPREAD, or COPY type=",type)
     if self.usage !="INPUT" and self.usage!="OUTPUT":
       SEP.util.err( "Distribution type must be INPUT or OUTPUT  usage="+self.usage)
     for i in range(len(self.isect_dict.keys())): self.naxis.append(0)
                                                                                       
  def check_file(self,dead):
     """Check to make sure provided files are logical"""
     if self.usage == "INPUT" or self.restart:
       if not os.path.isfile(self.name):
         SEP.util.err( "Input file["+self.name+"] doesn't exist")
       else: self.file=SEP.superset.superset(file=self.name)
     else:
       self.file=SEP.superset.superset(name=self.name)
       for i in range(self.axis-1):
         self.file.set_axis(i+1,1,0.,1.,"none","none")
       self.file.set_axis(self.axis,self.nsect,0.,1.,"none","none")
       self.file.write_file()
       
     if not self.restart:
       if self.file.dff != None:
         SEP.util.err("Input shouldn't be a distributed file unless restarting")
       else:
         self.par_file=SEP.superset.superset(supersetfile=self.file,
           name=self.name+"_par")
         self.par_file.create_dff(self.axis,self.nsect,self.type)
         self.par_file.write_file()
     else:                                                 
       if not os.path.isfile(self.name+"_par"):
         SEP.util.err( "Parallel file["+name+"_par] doesn't exist")
       else: 
         self.par_file=SEP.superset.superset(file=self.name+"_par")
         self.naxis=self.par_file.naxis_isect()
         mine = self.status.get_other_dict(self.status.get_by_status("finished"))
         mach_pair=self.mach.mach_map()
         for key,val in  mine.items(): self.mach_dict[key]=mach_pair[val]
