import vector
import types
import string
import os
import re
import SEP.sepfile
import SEP.util
import pwd
import SEP.datapath
import SEP.sep_vector
import SEP.spawn
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.2.27"
                                                                                


verb=0
temp_num=0

def solver_func(args,big_args):
  """Run a mathetical operation using Math

     args     - a list of arguments for Math
     big_args - Additional arguments when needing larger than 2GB files

  """
  args_to=["Solver_ops",big_args]
  args_to.extend(args)
  args_to.append("verb=y")
  tfile=SEP.util.temp_file("solver")
  SEP.spawn.run_wait(args_to,logfile=tfile)
  SEP.spawn.run_wait("rm -f %s"%tfile)
  
  
def math_func(args,big_args):
  """Run a mathetical operation using Math

     args     - a list of arguments for Math
     big_args - Additional arguments when needing larger than 2GB files

  """
  verbage=None
  if verb>2 : verbage=1
  args_to=["Math_base",big_args]
  args_to.extend(args)
  SEP.spawn.run_wait(args_to)
  
  
def rand_func(args,big_args):
  """Run the program  Noise to create a randome file

     args     - a list of arguments for Noise
     big_args - Additional arguments when needing larger than 2GB files

  """
  verbage=None
  if verb>2 : verbage=1
  args_to=["Noise",big_args]
  args_to.extend(args)
  SEP.spawn.run_wait(args_to)
  
  
class sep_vector(vector.vector,SEP.sepfile.sep_file):
  """ A class for out of core vectors in SEPlib file format"""

  def __init__(self,name=None,tag=None,vector=None,space_only=None,extra_args=" "):
     """Initialize a sep vector

        space_only - only initialze description don't wory about data
        extra_args - extra arguments that should be added when ever using
                      this vector (maybe needed for large files)

        Several different methods to initialize

        Initialze a vector from a file
          b=SEP.vector.sep_vector(tag) 
 
        Initialze a vector from another vector
          b=SEP.vector.sep_vector("aa",vector=a)
    
        Initialze an empty vector that we will define shape later
          b=SEP.vector.sep_vector("aa")

     """

     if not name: 
       if tag: self.name=tag
       else : SEP.util.err("Must initialize vector with a name")
     else: self.name=name
     self.extra_args=extra_args
     if extra_args: self.big_args=string.join(self.extra_args)
     else: self.big_args=" "
     self.e_size=None
     self.clones=0
     if tag:
       self._tag_init(tag)
     elif vector:
       self._vector_init(vector,space_only)
     else:
       SEP.sepfile.sep_file.__init__(self,name=self.name)
     self.space_only=space_only

  def add_extra_args(self,args):
    """Add extra args for programs  using this vector"""
    self.extra_args="%s %s"%(self.extra_args,args)

  def clone_space(self,name=None):
    """Clone just the space of a vector give it the name 'name'"""
    return self.clone(name=name,space_only=1)
  
  
  def _tag_init(self,tag):
    """Initialize a vector from a septag"""
    SEP.sepfile.sep_file.__init__(self,file=tag,name=self.name)

  def _vector_init(self,vector,space_only):
    """Initialize a sepvector from another vector"""
    arg=" "
    SEP.sepfile.sep_file.__init__(self,sepfile=vector,name=self.name)
    if self.headers:
      infile,err=self.header_par("in")
      self.set_header_par("in",infile)
    if self.grid:
      infile,err=self.grid_par("in")
      self.set_grid_par("in",infile)
    if not space_only: 
      if vector.space_only:self.create_data()
      else: self.copy_data(vector)

  def copy_data(self,vector):
    """Copy the data from one vector to another"""
    SEP.util.err("copy_data must be overwritten")
  def create_data(self,vector):
    """Create the data for a vector"""
    SEP.util.err("copy_data must be overwritten")

  def zero(self):
    """Zero a vector"""
    SEP.util.err("zero must be overwritten")

  def random(self):
     """Put random  numbers into vector"""
     args=["file=%s"%self.name]
     rand_func(args,self.big_args)

  def clean(self):
     """Remove a vector"""
     if verb >1: verbage=1
     else: verbage=None
     if not self.space_only: SEP.spawn.run_wait(["Rm3d",self.name],verb=verbage)



class sfloat(sep_vector):

  def __init__(self,name=None,tag=None,vector=None,space_only=None,extra_args=" "):
     """Initialize a sep vector composed of floats

        space_only - only initialze description don't wory about data
        extra_args - extra arguments that should be added when ever using
                      this vector (maybe needed for large files)

        Several different methods to initialize

        Initialze a vector from a file
          b=SEP.vector.sfloat(tag) 
 
        Initialze a vector from another vector
          b=SEP.vector.sfloat("aa",vector=a)
    
        Initialze an empty vector that we will define shape later
          b=SEP.vector.sfloat("aa")

     """
     sep_vector.__init__(self,name,tag,vector,space_only," ")

  def clone(self,space_only=None,name=None):
    """Clone a sep vector float 

        space_only - only clone the space
        name       - the name for the cloned vectror

    """
    if not name: name=self.name+".clone"+str(self.clones)
    self.clones=self.clones+1
    junk=sfloat(name,space_only=space_only,vector=self,extra_args=self.extra_args)
    return junk

  def load(self):
    """Load a sep vector file"""
    self.space_only=None

  def copy_data(self,vector):
     """Copy the data into this file"""
     SEP.spawn.run_wait("Cp %s %s "%(vector.name,self.name))
                                                                                


  def create_data(self):
     """out=0."""
     self.history.add_param("esize",4)
     self.history.add_param("in","-1")
     arg=["file1=%s"%self.name,"to_stdout=0",">/dev/null"]
     self.write_file()
     arg.append("exp='0.'")
     math_func(arg,self.big_args)
     self.space_only=None

  def zero(self):
     """Zero a vector"""
     if self.space_only:
       self.create_data()
     else:
       arg=["file1=%s"%self.name,"to_stdout=0",">/dev/null","exp='0.'"]
       math_func(arg,self.big_args)


  def scale(self,scale):
     """out=out*scale"""
     arg=["file1=%s"%self.name,"to_stdout=0",">/dev/null"]
     sc1=str(scale)
     if scale >= 0.: sc1="+"+sc1
     arg.append("exp='file1*%s'"%sc1)
     math_func(arg,self.big_args)

  def add(self,add):
     """out=out+add"""
     arg=["file1=%s"%self.name,"file2=%s"%add.name,"to_stdout=0",">/dev/null"]
     arg.append("exp='file1+file2'")
     math_func(arg,self.big_args)

  def scale_addscale(self,scale1,add,scale2):
     """out=out*scale1+scale2*add"""
     if type(scale1) is types.ComplexType: rscale1=scale1.real
     else: rscale1=scale1
     if type(scale2) is types.ComplexType: rscale2=scale2.real
     else: rscale2=scale2
     sc1=str(rscale1)
     sc2=str(rscale2)
     if rscale2 >= 0.: sc2="+"+sc2
     arg=["file1=%s"%self.name,"file2=%s"%add.name,"to_stdout=0",">/dev/null"]
     arg.append("exp='file1*%s%s*file2'"%(sc1,sc2))
     math_func(arg,self.big_args)

  def multiply(self,othervec):
     """Multiply a vector with another vector"""
     arg=["file1=%s"%self.name,"file2=%s"%othervec.name,"to_stdout=0",">/dev/null"]
     arg.append("exp='file1*file2'")
     math_func(arg,self.big_args)

  def dot(self,othervec):
     """Dot product vector with  another vector """
     if self.name==othervec.name: file2="file1"
     else: file2="file2"
     arg=["file1=%s"%self.name,"%s=%s"%(file2,othervec.name),"to_stdout=0","sum_axis=1"]
     arg.append("exp='file1*%s'"%file2)
     fle="/tmp/temp.%d"%os.getpid()
     arg.append(">"+fle)
     math_func(arg,self.big_args)
     par=SEP.par.sep_pars(file=fle)
     err2=1
     r,err1=par.par("sum_totr")
     if err1!=1 or err2!=1: 
       SEP.util.err("Trouble getting dot product results from  %s "%fle)
     SEP.spawn.run_wait(["rm",fle])
     return complex(float(r),0.)

  def size(self):
    """Return the size of the vector"""
    n=4
    i=1
    val,err=self.history_par("n%d"%i)
    while err==1:
      n=n*int(val)
      i=i+1
      val,err=self.history_par("n%d"%i)
    return n

#maybe should also inherit sepfile
class scmplx(sep_vector):

  def __init__(self,name=None,tag=None,vector=None,space_only=None,extra_args=" "):
     """Initialize a sep vector composed of complex numbers

        space_only - only initialze description don't wory about data
        extra_args - extra arguments that should be added when ever using
                      this vector (maybe needed for large files)

        Several different methods to initialize

        Initialze a vector from a file
          b=SEP.vector.scmplx(tag) 
 
        Initialze a vector from another vector
          b=SEP.vector.scmplx("aa",vector=a)
    
        Initialze an empty vector that we will define shape later
          b=SEP.vector.scmplx("aa")

     """
     sep_vector.__init__(self,name,tag,vector,space_only,extra_args)

  def load(self):
    """Load a sep vector file"""
    self.space_only=None

  def copy_data(self,vector):
     """Copy the data into this file"""
     SEP.spawn.run_wait("Cp %s %s "%(vector.name,self.name))

  def clone(self,space_only=None,name=None):
    """Clone a sep vector complex 

        space_only - only clone the space
        name       - the name for the cloned vectror

    """
    if not name: name=self.name+".clone"+str(self.clones)
    self.clones=self.clones+1
    if not name: name=self.name+".clone"+str(self.clones)
    self.clones=self.clones+1
    junk=scmplx(name,space_only=space_only,vector=self,extra_args=None) 
    return junk


  def create_data(self):
     """out=0."""
     self.history.add_param("esize",8)
     self.history.add_param("in","-1")
     self.write_file()
     arg=["file1=%s"%self.name,"op=zero",">/dev/null"]
     solver_func(arg,self.big_args)
     self.space_only=None

  def zero(self):
     """Zero a vector"""
     if self.space_only:
       self.create_data()
     else:
       arg=["file1=%s"%self.name,"op=zero",">/dev/null"]
       solver_func(arg,self.big_args)

  def scale(self,scale):
     """out=out*scale"""
     arg=["file1=%s"%self.name,"op=scale",">/dev/null"]
     arg.extend(["scale1_r='%f'"%scale])
     solver_func(arg,self.big_args)

  def add(self,add):
     """out=out+add"""
     arg=["file1=%s"%self.name,"file2=%s"%add.name,"op=add",">/dev/null"]
     solver_func(arg,self.big_args)

  def scale_addscale(self,scale1,add,scale2):
     """out=out*scale1+scale2*add"""
     if type(scale1) is types.ComplexType: 
        sc1r=scale1.real
        sc1i=scale1.imag
     else:
        sc1r=scale1
        sc1i=0.

     if type(scale2) is types.ComplexType: 
        sc2r=scale2.real
        sc2i=scale2.imag
     else:
        sc2r=scale2
        sc2i=0.
     if sc2i >= 0.: sc2i="+"+str(sc2i)
     if sc2r >= 0.: sc2r="+"+str(sc2r)
     arg=["file1=%s"%self.name,"file2=%s"%add.name]
     arg.extend(["op=scale_addscale",">/dev/null", "scale1_r=%s"%str(sc1r)])
     arg.append("scale1_i=%s"%str(sc1i))
     arg.append("scale2_r=%s"%str(sc2r))
     arg.append("scale2_i=%s"%str(sc2i))
     solver_func(arg,self.big_args)

  def multiply(self,othervec):
     """Multiply a vector with another vector"""
     arg=["file1=%s"%self.name,"file2=%s"%othervec.name,
       ">/dev/null","op=multiply"]
     solver_func(arg,self.big_args)

  def dot(self,othervec):
     """Dot product vector with  another vectory """
     arg=["file1=%s"%self.name,"op=dot"]
     if self.name!=othervec.name: arg.append("file2=%s"%othervec.name)
     else: arg.append("file2=%s"%self.name)
     fle="/tmp/temp.%d"%os.getpid()
     arg.append(">"+fle)
     solver_func(arg,self.big_args)
     f=open(fle)
     lines=f.readlines()
     r=re.compile("\s*DOT\s+RESULT\s*\(\s*(.+)\s*,\s*(.+)\s*\)")
     for line in lines:
       fnd=r.search(line)
       if fnd: 
         b=complex(float(fnd.group(1)),float(fnd.group(2)))
         SEP.spawn.run_wait(["rm",fle])
         return b 
     SEP.util.err("Trouble with dot product")

  def size(self):
    """Return the size of the vector"""
    n=8
    i=1
    val,err=self.history_par("n%d"%i)
    while err==1:
      n=n*int(val)
      i=i+1
      val,err=self.history_par("n%d"%i)
    return n
