import re
import string
import pwd
import sys
import os
import socket
import time
import SEP.par
import commands
from math import *
__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2004.12.2"
                                                                                


def min_max(n,o,d):
  """The minimum and maximum for an axis""" 
  n=int(n)
  o=float(o)
  d=float(d)
  end=o+d*(n-1)
  if end > o: return o,end
  else : return end,o 

def axis_same(label1,n1,o1,d1,label2,n2,o2,d2):
  """Check to see if the axes are the same""" 
  err=[]
  n1=int(n1)
  n2=int(n2)
  o1=float(o1)
  o2=float(o2)
  d1=float(d1)
  d2=float(d2)
  if n1 != n2: err.append("ns(%d,%d) not the same"%(n1,n2))
  if abs(o1-o2) > abs(d1)/10.: err.append("os(%f,%f) not the same"%(o1,o2))
  if abs(d1-d2) > abs(d1)/10.: err.append("ds(%f,%f) not the same"%(d1,d2))
  if err: SEP.util.err("Incompatibility between %s and %s\n%s"%(label1,
     label2,string.join(err,"\n")))

class format_file(SEP.par.sep_pars):
  """Extension of sep_par class that also knows about axes, in"""

  def __init__(self,file=None,name=None,sepfile=None):
    """Intialize a SEPlib format file

       Examples:

         file -  Initialize a format file from a file
            sfile.SEP.sepfile.format_file(file=file)
         name -  Intialize an empty format file
            sfile.SEP.sepfile.format_file(name="new.H")
         sepfile -  Intialize a format file from another format file structure
            sfile.SEP.sepfile.format_file(sepfile=sepfile)

    """
    SEP.par.sep_pars.__init__(self,file,name,sepfile)

  def axis(self,iax,ext=None):
     """Return axis (n,o,d,label,unit) in a format file

        Returns a given axis in a file  

        Examples:

         sfile.axis(1) returns n1,o1,d1,label1,unit1
         sefile.axis(1,'_grid') return n1_grid,o1_grid, etc

     """
     if ext == None: e=""
     else:  e=ext
     n=1
     o=0
     d=1
     unit="none"
     label="none"
     val,err = self.par("n"+str(iax)+e)
     if err == 1 : n= val
     val,err = self.par("o"+str(iax)+e)
     if err == 1 : o= val
     val,err = self.par("d"+str(iax)+e)
     if err == 1 : d= val
     val,err = self.par("label"+str(iax)+e)
     if err == 1 : label= val
     val,err = self.par("unit"+str(iax)+e)
     if err == 1 : unit= val
     return int(n),float(o),float(d),label,unit

  def set_axis(self,iax,n,o,d,label="",unit="",e=None):
    """Set axis in a format file description

       Example:
        sfile.set_axis(1,7,0.,1.) sets
          n1=7,o1=0. ,d1=1. label1="" unit1="" 
        sfile.set_axis(1,4,0.,2.,'none',e='_grid')
          n1_grid=4,o1_grid=0.,d1_grid=2.,label1_grid='none',unit1_grid=''

    """
    if e == None: suf=""
    else : suf=e
    self.add_param('n'+str(iax)+suf,n,type="int")
    self.add_param('o'+str(iax)+suf,o)
    self.add_param('d'+str(iax)+suf,d)
    self.add_param('label'+str(iax)+suf,label)
    self.add_param('unit'+str(iax)+suf,unit)

  def first_axis(self):
    """Return the first axis in file"""
    return 1

  def axis_min_max(self,iax):
    """Return the minimum and maximum of a given axis

       Example:
         n1=10 o1=0. d1=1.
       sfile=axis_min_max(1) will return 0.,9.
         
    """
    n,o,d,label,unit=self.axis(iax)
    return min_max(n,o,d)

  def size(self):
    """Return the size of a dataset

       Example:
         n1=10 n2=1 n3=6 esize=8
       sfile.size() returns 240

    """
    sz=int(self.elem_size())
    for i in range(self.first_axis(),self.ndims()+1): 
      sz=sz*int(self.axis(i)[0])
    return sz
      

  def ndims(self,suf=None,ibeg=1):
    """Number of dimensions in file
   
       suf if we are looking for n2_grid rather than n2 (suf=_grid)
       ibeg is the first axis that should exist (defaults to 1)

    """
    if suf == None: e=""
    else : e=suf
    i=ibeg
    iax=i; err=1; not1=0
    s="n"+str(iax)+e
    val,err = self.par(s)
    if err == 1:  not1=1
    while err ==1:
       s="n"+str(iax)+e
       val,err = self.par(s)
       if err == 1 and int(val) !=1: not1=iax
       iax=iax+1
    return not1

  def create_fake_in(self,name=None,file=None,host=None):
    """Create a fake data file """
    if name==None: nm=SEP.datapath.datafile(self.name,host)
    else: nm=name

    if file==None: fl=self.name
    else: fl=file

    self.add_param("in",nm)
    try: f = open (nm,"w")   #OPEN THE FILE
    except: SEP.util.err( 'Trouble opening file '+ nm)
    f.write(" ") 
    f.close()

class history(format_file):
  """Extension of the format_file class for SEPlib history file"""
  def __init__(self,file=None,name=None,sepfile=None):
    """See format_file initialization"""
    SEP.sepfile.format_file.__init__(self,file,name,sepfile)
  def elem_size(self):
     """Get the basic element size for the dataset"""
     val,err=self.par("esize")
     if err==0: val=4
     return val

class hff(format_file):
  """Extension of the format_file class """
  def __init__(self,file=None,name=None,sepfile=None):
    """See format_file initialization"""
    SEP.sepfile.format_file.__init__(self,file,name,sepfile)
    self.nkeys=0

  def set_nkey(self,nkey):
    """Set the number of keys in a dataset"""
    self.nkeys=nkey
    self.add_param("n1",nkey,"int")

  def elem_size(self):
    """Element size for dataset"""
    return 4

  def key(self,ikey):
     """Return a key from a hff file

        Example:
          headerkey 3 is described by 
           hdrkey3='alpha'  hdrtype3='scalar_float' hdrformat3='xdr_float'

          sfile.key (3) will return

           ['alpha','scalar_float','xdr_float']

     """
     list=("hdrkey","hdrtype","hdrfmt")
     out=[]
     for name in list:
       grab=name+str(ikey)
       val,err = self.par(grab)
       if err==1: SEP.util.err( grab+" doesn't exist in file")
       out.append(val)
     return out[0],out[1],out[2]

  def set_key(self,name,type,fmt,ik=None):
     """Set a key in a hff file

        Example:
         sfile.set_key("alpha","scalar_float","xdr_float",3) will set

           hdrkey3='alpha'  hdrtype3='scalar_float' hdrformat3='xdr_float'
          
          If ik not specified the number of keys will be incremented and
           ik will default to nkeys

     """
     if ik==None: ikey=self.nkeys+1
     else: ikey=ik
     if ikey > self.nkeys: self.set_nkey(ikey)
     if ikey== None: SEP.util.err("ikey must be specified" )
     if name== None: SEP.util.err("name must be specified" )
     if type== None: SEP.util.err("type must be specified") 
     if fmt== None: SEP.util.err("fmt must be specified" )
     self.add_param("hdrkey"+str(ikey),name)
     self.add_param("hdrtype"+str(ikey),type)
     self.add_param("hdrmt"+str(ikey),fmt)


class gff(format_file):
  """Extension of the format_file class to handle a grid format file"""
  def __init__(self,file=None,name=None,sepfile=None):
    SEP.sepfile.format_file.__init__(self,file,name,sepfile)

  def axis(self,iax):
    """Return a given axis see SEP.sepfile.format_file.axis"""
    if iax ==1 : SEP.util.err("Can't request first axis from a grid")
    return format_file.axis(self,iax,"_grid")

  def elem_size(self):
    """Eleement size for dataset"""
    return 4

  def first_axis(self):
    """First axis in file"""
    return 2 #shouldn't this be 2

  def set_axis(self,iax,ext,n,o,d,label="",unit=""):
    """Set a given axis see SEP.sepfile.format_file.set_axis"""
    if iax ==1 : SEP.util.err("Can't set first axis from a grid")
    SEP.sepfile.format_file.set_axis(self,iax,n,o,d,label,unit,"_grid")
 
  def ndims(self):
    """Return the number of dimensions in a dataset"""
    return format_file.ndims(self,"_grid",2)


class sep_file:
  """A class for dealing with a SEPlib dataset

     Can handle regular, header, and grided datasets

     Can read and write files

   """

  def __init__(self,file=None,sepfile=None,name=None,fmt=None,remote=None):
    """Initialize a SEPlib file
 
       Initialize by specifying one or of the initialized parameters

       Initialize by a  (file), another sepfile (sepfile),
         an empty file (name and fmt), or a remote file (remote)

       Examples: 
         sfile=SEP.sepfile.sep_file(file="a.H")
         sfile=SEP.sepfile.sep_file(sepfile=sfile2)
         sfile=SEP.sepfile.sep_file(name="a.H",fmt="REGULAR")
         sfile=SEP.sepfile.sep_file(name="a.H",fmt="HEADERS")
         sfile=SEP.sepfile.sep_file(name="a.H",fmt="GRID")
         sfile=SEP.sepfile.sep_file(remote="sep201:a.H")
 

    """
    self.grid=None
    self.headers=None
    if remote!=None: self._remote_init(remote)
    elif file != None: self._file_init(file) 
    elif sepfile != None: 
      if name !=  None: self._sep_init(sepfile,name)
      else:
        SEP.util.err( "Must provide the name if initializing from another sepfile")
    elif name !=None: self._name_init(name,fmt)
    else:
      SEP.util.err( "Must initialize a sep_file with a file or another sep_file")
    self.name=self.history.name


  def _remote_init(self,remote):
    """Initialize a file from a remote location"""
    pts=remote.split(":")
    if len(pts)!=2: SEP.util.err( "Expecting remote location of the form host:file")
    mach=pts[0]
    file=pts[1]
    temp_file=SEP.util.temp_file("temp_rep")
    SEP.spawn.run_wait(["rcp "+mach+":"+file+" "+temp_file])
    self.history=SEP.sepfile.history(temp_file)
    list={'hff':"@@",'gff':"@@@@"}
    cp_com=[]
    rm_com=["rm "+temp_file]
    for part,ext in list.items():
      val,err=self.history.par(part)       
      if err ==1 and val!="-1": 
        rm_com.append("rm "+temp_file+ext)
        cp_com.append("rcp "+mach+":"+val+" "+temp_file+ext)
    for com in cp_com: SEP.spawn.run_wait(string.split(com))
    val,err=self.history_par("hff")
    if err==1 and val != "-1":
      self.hff=SEP.sepfile.hff(temp_file+"@@")
      val,err=self.history_par("gff")
      if err==1 and val != "-1":
        self.gff=SEP.sepfile.gff(temp_file+"@@@@")
        self.file_format="GRID"
      else:  self.file_format="HEADERS"
    else: self.file_format="REGULAR"
    for com in rm_com: SEP.spawn.run_wait(string.split(com))

  def _name_init(self,name,fmt):
    """Initialize a blank file"""
    if fmt==None: self.file_format="REGULAR"
    else: self.file_format=fmt
    if self.file_format != "REGULAR" and  self.file_format != "HEADERS" and  self.file_format != "GRID":
     SEP.util.err( fmt+"= format must be REGULAR,HEADERS,or GRID")
    self.history=SEP.sepfile.history(name=name)
    self.history.set_axis(1,1,0.,1.,"none","none")
    if self.file_format=="REGULAR": 
      self.history.set_axis(2,1,0.,1.,"none","none")
    else:
      self.headers=SEP.sepfile.hff(name=name+"@@")
      self.headers.set_axis(2,1,0.,1.,"none","none")
      self.headers.set_key("none","scalar_float","xdr_float")
      if self.file_format=="GRID":
        self.grid=SEP.sepfile.gff(name=name+"@@@@")
        self.grid.set_axis(2,1,0.,1.,"none","none")

  def _sep_init(self,sepfile,name):
    """Initalize a sep_file object from a sep_file object"""
    self.history=SEP.sepfile.history(sepfile=sepfile.history,name=name)
    self.file_format=sepfile.file_format
    if sepfile.file_format != "REGULAR":
      self.headers=SEP.sepfile.hff(sepfile=sepfile.header,name=name+"@@")
      if sepfile.file_format == "GRID":
        self.grid=SEP.sepfile.gff(sepfile=sepfile.grid,name=name+"@@@@")

  def _file_init(self,file):
    """Initalize a sep_file object from a file"""
    try: f = open (file)   #OPEN THE FILE
    except: SEP.util.err( 'Trouble opening file:'+ file)
    f.close()
    self.history=SEP.sepfile.history(file=file)
    val, err = self.history.par("hff")
    if err == 1 and val != "-1":
      self.headers=SEP.sepfile.hff(file=val)
      val, err = self.history.par("gff")
      if err ==1 and val != "-1":
        self.grid=SEP.sepfile.gff(file=val)
        self.file_format="GRID"
      else: self.file_format="HEADERS"
    else: self.file_format="REGULAR"

  def history_par(self,par,error=None):
     """Get parameter from history file"""
     return self.history.par(par,error=error)

  def header_par(self,par,error=None):
     """Get parameter from hff file"""
     if self.file_format == "REGULAR": return None,-1
     return self.headers.par(par,error=error)

  def grid_par(self,par,error=None):
     """Get parameter from gff file"""
     if self.file_format != "GRID": return None,-1
     return self.grid.par(par,error=error)

  def add_history_par(self,par,value,type="str"):
     """Set parameter in history file"""
     return self.history.add_param(par,value,type)

  def add_header_par(self,par,value,type="str"):
     """Set parameter in hff file"""
     if self.file_format == "REGULAR": return None,-1
     return self.headers.add_param(par,value,type)

  def add_grid_par(self,par,value,type="str"):
     """Set parameter in gff file"""
     if self.file_format != "GRID": return None,-1
     return self.grid.add_param(par,value,type)

  def axis(self,iax):
     """Set axis for a SEP dataset

        SEE: SEP.sepfile.format_file.axis

     """
     if self.file_format == "GRID" and iax != 1:  
       return self.grid.axis(iax)
     elif self.file_format == "HEADERS" and iax !=1:  
       return self.headers.axis(iax)
     else :  
       return self.history.axis(iax)

  def write_file(self,base=None):
    """Write a seplib file to disk"""
    if self.file_format!="REGULAR":
      if base == None: self.headers.write_pars(self.headers.name)
      else:
        nmh,val=self.history.par("hff")
        self.history.add_param("hff",base+"@@")
        self.headers.write_pars(base+"@@")
        if self.file_format=="GRID":
          if base == None: self.grid.write_pars(self.headers.name)
          else:
            nmg,val=self.history.par("gff")
            self.history.add_param("gff",base+"@@@@")
            self.grid.write_pars(base+"@@@@")
    if base == None: self.history.write_pars( self.history.name)
    else: self.history.write_pars(base)
    if self.file_format  != "REGULAR" and base != None:
      self.history.add_param("hff",nmh)
      if self.file_format=="GRID" : self.grid.add_param("gff",nmg)

  def set_axis(self,iax,n,o,d,label="",unit=""):
     """Set an axis in a seplib dataset

        SEE: SEP.sepfile.format_file.set_axis

     """
     if self.file_format == "GRID" and iax!=1:  
       return self.grid.set_axis(iax,n,o,d,label,unit)
     elif self.file_format == "HEADERS" and iax!=1:  
       return self.headers.set_axis(iax,n,o,d,label,unit)
     else :  return self.history.set_axis(iax,n,o,d,label,unit)

  def axis_min_max(self,iax):
     """Return axis minimum and maximum 

        SEE: SEP.sepfile.format_file.axis_min_max

     """
     if self.file_format == "GRID" and iax!=1:  return self.grid.axis_min_max(iax)
     elif self.file_format == "HEADERS" and iax!=1:  return self.headers.axis_min_max(iax)
     else : return self.history.axis_min_max(iax)
     
  def create_fake(self,host=None):
    """Open up 0 length files and set in="""
    self.history.create_fake_in(host=host)
    if self.headers != None: self.headers.create_fake_in(host=host)
    if self.grid != None: self.grid.create_fake_in(host=host)

  def remove(self,add="",verb=None):
    """Remove the format and binary files from disk if they exist"""
    parts=[self.history,self.headers,self.grid]
    list=[]
    for part in parts:
      if part != None:
        if os.path.isfile(part.name): list.append("rm %s "%add+part.name)
        val,err = part.par("in")
        if err==1: 
          if os.path.isfile(val): list.append("rm %s %s"%(add,val))
    for com in list: SEP.spawn.run_wait(string.split(com),verb=verb)
     

  def send(self,host,file,binary=None):
     """Transfer a sepfile to another host

         transfer to host (host) with the filename (file) 

         If binary is defined also transfer binary associated with 
         format files

     """
     temp=SEP.util.temp_file("temp_rep")
     tempstruct=SEP.sepfile.sep_file(sepfile=self,name=temp)
     copy_list={}
     copy_list[tempstruct.history.name]=file
     if binary:
         bin=SEP.datapath.datafile(file,host)
         val,err = self.history.par("in")
         if err == 1:  
           copy_list[val]=bin
           tempstruct.history.add_param("in",bin)
     erase_com=["rm "+temp]
     if tempstruct.headers != None:
        erase_com.append("rm "+temp+"@@")
        copy_list[tempstruct.headers.name]=file+"@@"
        if binary:
           bin=SEP.datapath.datafile(file+"@@",host)
           val,err = self.headers.par("in")
           if err == 1:  
             copy_list[val]=bin
             tempstruct.headers.add_param("in",bin)
        tempstruct.history.add_param("hff",file+"@@")
     if tempstruct.grid != None:
        erase_com.append("rm "+temp+"@@@@")
        copy_list[tempstruct.grid.name]=file+"@@@@"
        if binary:
           bin=SEP.datapath.datafile(file+"@@@@",host)
           val,err = self.grid.par("in")
           if err == 1:  
             copy_list[val]=bin
             tempstruct.grid.add_param("in",bin)
        tempstruct.history.add_param("gff",file+"@@@@")
     tempstruct.write_file(temp)
     copy_com=[]
     for myl,r in copy_list.items():
       copy_com.append("rcp "+myl+" "+host+":"+r)
     for com in copy_com: SEP.spawn.run_wait(string.split(com))
     for com in erase_com: SEP.spawn.run_wait(string.split(com))

  def size(self):
     """Return the size of a SEPlib3d dataset
        
        If grided  dataset returns the size of the grid
        If headers dataset returns the size of the headers
        If regular dataset returns the size of the regular dataset

        SEE: SEP.sepfile.format_file for more information

     """
     if self.file_format == "GRID":  return self.grid.size()*self.axis(1)[0]
     elif self.file_format == "HEADERS":  return self.headers.size()*int(self.history.axis(1)[0])
     elif self.file_format == "REGULAR":  return self.history.size()*int(self.history.axis(1)[0])
     else: SEP.util.err("File format not specified")

  def ndims(self):
     """Return the number of dimensions in a SEPlib3d dataset

        SEE: SEP.sepfile.format_file for more information

     """
     if self.file_format == "GRID":  return self.grid.ndims()
     elif self.file_format == "HEADERS":  return self.headers.ndims()
     elif self.file_format == "REGULAR":  return self.history.ndims()
     else: SEP.util.err("File format not specified")

