import SEP.util
import sys,string,re,commands,os
import SEP.par
import SEP.spawn


def copy_list(list,dir,dir_base):
  for file in list: 
    if os.path.basename(file) == file: file=dir_base+os.sep+file
    SEP.spawn.run_wait(["cp",file,dir])

class wei_component:
  def __init__(self,modname,operator,init,code,extra=None):
    self.op=operator
    self.init=init
    self.mod_name=modname
    self.type=None
    self.name=operator
    self.extra=extra
    self.code=code
  def component_check(self,component):
    return 1
  def copy_code(self,dir,dir_base):
    list=[self.code]
    if self.extra: list.extend(self.extra)
    copy_list(list,dir,dir_base)
    
class fk(wei_component):
  def __init__(self,modname,operator,init,code,extra=None):
     wei_component.__init__(self,modname,operator,init,code,extra)
     self.type="fk"
class fx(wei_component):
  def __init__(self,modname,operator,init,code,extra=None):
     wei_component.__init__(self,modname,operator,init,code,extra)
     self.type="fx"
class ig(wei_component):
  def __init__(self,modname,operator,init,code,extra=None):
     wei_component.__init__(self,modname,operator,init,code,extra)
     self.type="ig"
class sl(wei_component):
  def __init__(self,modname,operator,init,code,extra=None):
     wei_component.__init__(self,modname,operator,init,code,extra)
     self.type="sl"
class wc(wei_component):
  def __init__(self,modname,operator,init,code,fks,fxs,extra=None):
     wei_component.__init__(self,modname,operator,init,code,extra)
     self.type="wc"
     self.fks=fks
     self.fxs=fxs
  def component_check(self,components):
    if 0== self.fks.count(components["fk"].name):
      SEP.util.err("%s doesn't understand fk=%s, understands fks=[%s]"%(self.name,
        components["fk"].name,string.join(self.fks)))
    if 0== self.fxs.count(components["fx"].name):
      SEP.util.err("%s doesn't understand fx=%s, understands fxs=[%s]"%(self.name,
        components["fx"].name,string.join(self.fxs)))
    return 1
class down(wei_component):
  def __init__(self,modname,operator,init,code,modes,extra=None):
     wei_component.__init__(self,modname,operator,init,code,extra=None)
     self.modes=modes
     self.type="down"

#should be moved and or replaced
class simple_package:
  def __init__(self,name,libdirs,libs,includedirs=None):
    self.libdirs=libdirs
    self.libs=libs
    self.includedirs=includedirs
    self.name=name
  def return_lib_line(self):
    lineout=""
    for dir in self.libdirs: lineout+=" -L%s "%dir
    for lib in self.libs  : lineout+=" -l%s "%lib
    return lineout
  def return_inc_line(self):
    lineout=""
    if  self.includedirs:
      for inc in self.includedirs  : lineout+=" -I%s "%inc
    return lineout
    


class wei_code:
  """Wei code object"""
  def __init__(self,components,progname,base_dir,base_code,packages,build_dir):
    """Initialize wei code object"""
    self.comp_name=["sl","wc","fk","fx","ig"]
    self.op_name="down"
    self.packages=packages
    self.base_dir=base_dir
    self.build_dir=build_dir
    names=[self.op_name]
    self.base_code=base_code
    for val in self.comp_name: names.append(val)
    self.components={}
    for elem in components: self.components[elem.type]=elem
    for name in names:  
      if  not self.components.has_key(name):
        SEP.util.err("wei_basic not initialized with %s",name)
    for comp in self.components.values():
       comp.component_check(self.components)
    binbase=re.compile("(.+).x$").search(progname)
    self.name=progname
    if not binbase: SEP.util.err("Expecting program(%s) to end with .x"%progname)
    progbase=binbase.group(1)
    self.basename=os.path.basename(progbase)
    binS=re.compile("(.+)"+self.basename).search(progbase)
    if binS: self.bindir=binS.group(1)
    else: self.bindir=os.getcwd()

  def copy_dir(self,dir):
    """Copy the code for a wei program"""
    SEP.spawn.run_wait(["mkdir","-p",dir])
    for comp in self.components.values(): comp.copy_code(dir,self.base_dir)
    copy_list(self.base_code,dir,self.base_dir)
     
  def build_main(self,dir):
     """Build the main program"""
     prog="%s%s%s.f90"%(dir,os.sep,self.basename)
     try: f = open(prog,"w")
     except: SEP.util.err("Trouble opening %s for writing"%prog)
     f.write("Program %s\n"%self.basename)
     f.write("  use sep\n")
     for comp in self.components.values(): f.write("  use %s\n"%comp.mod_name)
     f.write("  implicit none\n")
     f.write("  integer             :: ierr \n")
     f.write("  logical             :: adj,add \n")
     f.write("  character(len=128)  :: R,D \n\n\n");

     f.write("  call sep_init(SOURCE)\n")
     f.write("  call getch_add_string(\"head=/dev/null\")\n")
     f.write("  call getch_add_string(\"noheader=y\")\n")
     f.write("  call from_param(\"adj\",adj,.true.)\n")
     f.write("  call from_param(\"add\",add,.false.)\n")
     f.write("  R='R'\n")
     f.write("  D='D'\n")


     f.write("call %s ("%self.components[self.op_name].init)
     list=[]
     for comp in self.comp_name:
       list.append("%sin=%s"%(self.components[comp].type,self.components[comp].init))
     f.write(string.join(list,",&\n      "))
     f.write(")\n\n\n")

     f.write("ierr=  %s (adj,add,R,D, "%self.components[self.op_name].op)
     list=[]
     for comp in self.comp_name:
       list.append("%sop=%s"%(self.components[comp].type,self.components[comp].op))
     f.write(string.join(list,",&\n      "))

     f.write("      )\ncall sep_end_prog()\n")
     f.write("call exit(0)\n\n")
     f.write("end program\n")
     f.close()

     
  def build_make(self,dir,options):
     """Build the makefile"""
     make="%s%s%s"%(dir,os.sep,"Makefile")
     try: f = open(make,"w")
     except: 
       SEP.util.err("Trouble opening makefile %s \n"%make)
     f.write("include ${SEPINC}/SEP.top\n")
   
     f.write("BINDIR=%s\n"%self.bindir)
     if options:
       for key,val in options.items(): 
         f.write("%s=%s\n"%(key,val))

     f.write("UF90INCLUDES= ")
     for pack in self.packages: 
       f.write(pack.return_inc_line())

     f.write("\nUF90LIBS=")
     for pack in self.packages: 
       f.write(pack.return_lib_line())

     f.write("\n\n\nall:${BINDIR}/%s.x\n\n\n"%self.basename)
     f.write("include ${SEPINC}/SEP.bottom\n")
     f.close()

  def build_source(self,options=None):
    """Generate a makefile and the code the build an executable"""
    self.copy_dir(self.build_dir)
    self.build_main(self.build_dir)
    self.build_make(self.build_dir,options)
  
  def build(self,options=None): 
    self.build_source(options) 
    SEP.spawn.run_wait(["cd ",self.build_dir,"; make all"],logfile=self.build_dir+os.sep+"log")


base_code=["wei_basics.F90","wei_external.F90","wei_imaging.F90","wei_kmap.F90",
 "wei_fft.F90","wei_data.F90","wei_mem.F90","wei_flags.F90","wei_image.F90",
 "wei_run.F90","wei_par.F90","wei_verb.F90", "wei_wavespace.F90","wei_slow.F90",
"wei_xmap.F90","wei_path.F90","wei_timers.F90"]
  

cam=fk("wei_cam","weicam","weicam_init","wei_cam.F90",["wei_fk_share.F90"])
wem=fk("wei_wem","weiwem","weiwem_init","wei_wem.F90",["wei_fk_share.F90"])
ssf=fx("wei_ssf","weissf","weissf_init","wei_ssf.F90")
hcig=ig("wei_cig","weihcig","weihcig_init","wei_cig.F90")
pcig=ig("wei_pcig","weipcig","weipcig_init","wei_pcig.F90")
sll=sl("wei_sll","weisll","weisll_init","wei_sll.F90",["lloyd.r90"])
pspi=wc("wei_wc_mod","weikxmign_mid","weikxmign_init","wei_pspi.F90",
   ["weicam","weiwem"],["weissf"])
migz=down("wei_mig_base","weimigz","weimigz_init","wei_mig_base.F90",
  ["forward","adjoint"])
