import SEP.options
import SEP.par
import SEP.lloyd
import SEP.util
import SEP.sepfile

__author__ = "Robert G. Clapp"
__email__ = "bob@sep.stanford.edu"
__version = "2005.2.1"
                                                                                


class wtransp_interp(SEP.options.par_group):
  """A transpose/interpolation operation that is useful for wei"""
  def __init__(self,main_par):
    SEP.options.par_group.__init__(self,"WEI.TI")
    self.add_doc_params()
    self.main=main_par
  def add_doc_params(self):
    """Parameters for the transp wei needs"""
    self.add_doc_param("maxsize",100,"Approximate memory to use")
    self.add_doc_param("interp_type",1,"Interpolation")
  def build_check_params(self):
    """Build the commands for the program"""
    vel=self.main.param("velocity")
    oz=float(self.main.param("oz"))
    dz=float(self.main.param("dz"))
    nz=int(self.main.param("nz"))
    vout="mig_%s"%vel
    self.add_param("vel_mig",vout)
    self.command="<%s %s/Interp maxsize=%s type=%s  "%(vel,
      SEP.paths.sepbindir,self.param("maxsize"),self.param("interp_type"))
    self.command="%s n1out=%d o1out=%f d1out=%f |%s/Transp"%(self.command,
      nz,oz,dz,SEP.paths.sepbindir)
    self.command="%s reshape=1,3 maxsize=%s >%s plane=12 "%(self.command,
      self.param("maxsize",vout),vout)
  def run_job(self,add_args=None):
    """Run the Interp|Transp program"""
    SEP.spawn.run_wait(self.command,verb=self.param("verb",error=None)) 


class lloyd_wei(SEP.lloyd.lloyd):
  """Lloyd for WEI"""
  def __init__(self,main_args):
    """Initialze the lloyd operator WEI migration"""
    self.main=main_args
    SEP.lloyd.lloyd.__init__(self,"WEI.LLOYD")
  def add_doc_params(self):
    """Add the documentation for Lloyd"""
    SEP.lloyd.lloyd.add_doc_params(self)
    self.del_doc_param("stdin")
    self.del_doc_param("stdout")
    self.del_doc_param("vel_map")
  def build_check_params(self):
    """Build commands for Lloyd"""
    vel="mig_%s"%self.main.param("velocity")
    schoice="ref_%s"%vel
    self.add_param("stdin",vel)
    self.add_param("stdout",schoice)
    self.add_param("schoice",schoice)
    SEP.lloyd.lloyd.build_check_params(self)


class parallel_wei(SEP.wei.wei_par):
  def __init__(self,wei_par,par_par,main_args):
      SEP.wei.wei_par.__init__(self,wei_par,par_par)
      self.main=main_args
  def build_check_params(self):
    self.add_param("restart",self.main.param("restart"))
    SEP.wei.wei_par.build_check_params(self)
  def run_job(self,restart,par=None):
    """Run the job (in this case migrate)"""
    self.migrate(restart)

class off_mig(SEP.wei.mig_off_par):
  def __init__(self,main_args):
     SEP.wei.mig_off_par.__init__(self,"WEI.MIG")
     self.main=main_args
  def add_doc_params(self):
     SEP.wei.mig_off_par.add_doc_params(self)
     self.del_doc_param("az__n")
     self.del_doc_param("az__o")
     self.del_doc_param("az__d")
     self.del_doc_param("program")
     self.del_doc_param("S")
     self.del_doc_param("schoice")
     self.del_doc_param("D")
     self.del_doc_param("R")
  def build_check_params(self):
    """Build the commands for the program"""
    self.add_param("program",self.main.param("program"))
    vel=self.main.param("velocity")
    oz=float(self.main.param("oz"))
    dz=float(self.main.param("dz"))
    nz=int(self.main.param("nz"))
    self.add_param("az__n",nz)
    self.add_param("az__o",oz)
    self.add_param("az__d",dz)
    self.velocity=self.main.param("velocity")
    self.add_param("schoice","ref_mig_%s"%self.velocity)
    self.add_param("S","mig_%s"%self.velocity)
    self.add_param("R",self.main.param("R"))
    self.add_param("D",self.main.param("D"))
    SEP.wei.mig_off_par.build_check_params(self)

class datuming(SEP.wei.datum_par):
  def __init__(self,main_arg):
     self.main=main_arg
     SEP.wei.datum_par.__init__(self,"WEI.DATUM")
  def add_doc_params(self):
     SEP.wei.datum_par.add_doc_params(self)
     self.del_doc_param("az__n")
     self.del_doc_param("az__o")
     self.del_doc_param("az__d")
     self.del_doc_param("program")
     self.del_doc_param("S")
     self.del_doc_param("schoice")
     self.del_doc_param("D")
     self.del_doc_param("DOWN")
  def build_check_params(self):
    """Build the commands for the program"""
    self.add_param("program",self.main.param("program"))
    vel=self.main.param("velocity")
    oz=float(self.main.param("oz"))
    dz=float(self.main.param("dz"))
    nz=int(self.main.param("nz"))
    self.add_param("az__n",nz)
    self.add_param("az__o",oz)
    self.add_param("az__d",dz)
    self.velocity=self.main.param("velocity")
    self.add_param("schoice","ref_mig_%s"%self.velocity)
    self.add_param("S","mig_%s"%self.velocity)
    self.add_param("DOWN",self.main.param("DOWN"))
    self.add_param("D",self.main.param("D"))
    SEP.wei.mig_off_par.build_check_params(self)

   
    
    
class off_mig_ph(SEP.wei.mig_ph_par):
  def __init__(self,main_arg):
     self.main=main_arg
     SEP.wei.mig_ph_par.__init__(self,"WEI.MIG")
  def add_doc_params(self):
     SEP.wei.mig_ph_par.add_doc_params(self)
     self.del_doc_param("program")
     self.del_doc_param("az__n")
     self.del_doc_param("az__o")
     self.del_doc_param("az__d")
     self.del_doc_param("D")
     self.del_doc_param("R")
     self.del_doc_param("S")
     self.del_doc_param("schoice")
  def build_check_params(self):
    """Build the commands for the program"""
    self.add_param("program",self.main.param("program"))
    vel=self.main.param("velocity")
    oz=float(self.main.param("oz"))
    dz=float(self.main.param("dz"))
    nz=int(self.main.param("nz"))
    self.add_param("az__n",nz)
    self.add_param("az__o",oz)
    self.add_param("az__d",dz)
    self.add_param("R",self.main.param("R"))
    self.add_param("D",self.main.param("D"))
    self.velocity=self.main.param("velocity")
    self.add_param("schoice","ref_mig_%s"%self.velocity)
    self.add_param("S","mig_%s"%self.velocity)
    SEP.wei.mig_ph_par.build_check_params(self)

class mig_main(SEP.options.par_group):
  """The migration """
  def __init__(self):
    SEP.options.par_group.__init__(self,"WEI.MAIN",is_job=None)
    self.add_doc_params()
  def add_doc_params(self):
    """Parameters for the transp wei needs"""
    self.add_doc_param("nz",doc="Number of depth steps")
    self.add_doc_param("oz",doc="First depth")
    self.add_doc_param("D",doc="Data (mx,my,hx,hy,w)")
    self.add_doc_param("R",doc="Image (mx,my,hx/px,hy/py,z")
    self.add_doc_param("dz",doc="Sampling in depth")
    self.add_doc_param("velocity",doc="Velocity Model")
    self.add_doc_param("restart",doc="Whether or not to restart",required=None)
    self.add_doc_param("program",doc="WEI program to use")
