#!/usr/bin/env python
import SEP.util
import sys,string,re,commands,os
import SEP.operator
import SEP.parfile
import SEP.parjob
import SEP.par
import SEP.sep_vector
import SEP.vector



class wei_operator(SEP.operator.oc_par):
  """Class that treats the migration as an operator"""
  def __init__(self,prog,pars,**kw):
     pars.add_dictionary(kw,"par_")
     par_par=pars.extract_del_pars("par_")

     par_par.default_val("mfile","mfile")
     par_par.default_val("njobs",2000)
     restart=par_par.logic("restart")
     logdir=par_par.param("logdir","wei_log")
     name=par_par.param("name","wei_run")
     par_par.default_val("restart_com","restart=y")

     self.pars=pars
     self.check_wavespace()
     slowS,schoice=self.slow_sep()
     isect_dict,sect_par,order=self.build_dicts()
     parfiles={}
     self.program=prog
     parfiles["schoice"]=SEP.parfile.copy(name=schoice.name,tag="schoice=",
        isect_dict=isect_dict,usage="INPUT",type="COPY",logdir=logdir)
     parfiles["slow"]=SEP.parfile.copy(name=slowS.name,tag="S=",
        isect_dict=isect_dict,usage="INPUT",type="COPY",logdir=logdir)
     parfiles["range"]=SEP.parfile.distribute(name="data",tag="D=",
        isect_dict=isect_dict,space_only=1,dff_axis=5,logdir=logdir,
       nblock=[len(isect_dict.keys())])
     if pars.param("nfs_image",0,delete=1)==1:
       nmax=pars.param("nmax_procs",None,delete=1)
       dpath=pars.param("share_datapath",None,delete=1)
       parfiles["domain"]=SEP.parfile.copy_share(name="model",tag="R=",logdir=logdir,
         isect_dict=isect_dict,space_only=1,type="COPY",
         reuse_par="add=y",nmax_procs=nmax, share_datapath=dpath)
     else:
       parfiles["domain"]=SEP.parfile.copy(name="model",tag="R=",logdir=logdir,
          isect_dict=isect_dict,space_only=1,type="COPY",reuse_par="add=y")
     if pars.par("wave_out")[1]==1:
       parfiles["wave_out"]=SEP.parfile.distribute(
         name=pars.param("wave_out",delete=1),tag="wave_out=",
        isect_dict=isect_dict,dff_axis=5,logdir=logdir,
        usage="OUTPUT", nblock=[len(isect_dict.keys())])
     model=self.create_model("mod_space")
     data=self.create_data("dat_space")

     parjob=SEP.parjob.parjob(files=parfiles,sect_pars=sect_par,
      program=self.program, order=order, **par_par.return_dict())

     if not name: name=self.program
     SEP.operator.oc_par.__init__(self,name,
       model,data,parjob,pars=pars,model_tag="R=",data_tag="D=",restart=restart)

     self.model_check={}
     self.data_check={}
     self.verb=pars.logic("verb")


  def build_dicts(self):
    """Return  the dictionary for both the files and for the parameters"""
    isect_dict={}
    par_dict={}
    nw=int(self.pars.par("aw__n")[0])
    ow=float(self.pars.par("aw__o")[0])
    dw=float(self.pars.par("aw__d")[0])
    order=[]
    for i in range(nw): 
       key=str(i)
       order.append(key)
       isect_dict[key]=i
       par_dict[key]=SEP.par.sep_pars(name=str(i))
       o=ow+dw*i
       par_dict[key].add_param("aw__n",1)
       par_dict[key].add_param("aw__o",o)
       par_dict[key].add_param("aw__d",dw)
    return isect_dict,par_dict,order

  def add_params(self,pars):
    """Add and check parameters"""
    SEP.operator.oc.add_params(self,pars)
    self.check_slow(slow)

  def slow_sep(self):
    """Check to make sure the slowness make sense"""
    val,err = self.pars.par("S")
    if err !=1: SEP.util.err("Expecting slowness in the form S=file.H")
    slow=SEP.sepfile.sep_file(file=val)
    val,err = self.pars.par("schoice")
    if err !=1: SEP.util.err("Expecting ref slowness in schoice=file.H")
    schoice=SEP.sepfile.sep_file(file=val)

    if schoice.headers or schoice.grid:
      SEP.util.err("Expecting schoice to be a regular SEPlib file")

    if slow.headers or slow.grid:
      SEP.util.err("Expecting slowness to be a regular SEPlib file")

    if schoice.ndims()!=3: SEP.util.err("Expecting schoice to be a 3-D file nref,2,nz")
    if slow.ndims()!=3: SEP.util.err("Expecting slowness to be a 3-D file x,y,z")
    if slow.history_par("n1")[0] > 1:
       b,e=slow.axis_min_max(1)
       b1=self.pars.param("wave_x_min",delete=1)
       e1=self.pars.param("wave_x_max",delete=1)
       if b >b1: SEP.util.msg("Warning: Slowness axis 1 will be extended (neg) by %f "%(b-b1))
       if e <e1: SEP.util.msg("Warning: Slowness axis 1 will be extended (pos) by %f "%(e1-e))
    n3_s,oz,dz,label,unit=schoice.history.axis(3)
    n3_c=slow.history_par("n3",error=1)[0]
    if int(n3_s) != int(n3_c): 
       SEP.util.err("Expecting slowness and schoice to have identical n3")

    if n3_s!=int(self.pars.par("az__n")[0]):
      SEP.util.err("Number of samples slowness (axis 3) not the same")
    if oz!=float(self.pars.par("az__o")[0]):
      SEP.util.err("First sample of slowness (axis 3) not the same")
    if dz!=float(self.pars.par("az__d")[0]):
      SEP.util.err("Sampling of slowness (axis 3) not the same")


    if slow.history_par("n2")[0] > 1:
       b,e=slow.axis_min_max(2)
       b1=self.pars.param("wave_y_min",delete=1)
       e1=self.pars.param("wave_y_max",delete=1)
       if b >b1: SEP.util.msg("Warning: Slowness axis 2 will be extended (neg) by %f "%(b-b1))
       if e <e1: SEP.util.msg("Warning: Slowness axis 2 will be extended (pos) by %f "%(e1-e))

       return slow,schoice

  def check_model(self,model):
    """Check to see if the model makes sense"""
    if self.model_check.has_key(model.name): return
    if model.headers or model.grid:
      SEP.util.err("Model %s should be a regular dataset"%model.name)
    self.axis_check("Model",model,1,"mx")
    self.axis_check("Model",model,2,"my")
    self.axis_check("Model",model,3,"hx")
    self.axis_check("Model",model,4,"hy")
    self.axis_check("Model",model,5,"z_")
    self.model_check[model.name]=1

  def axis_check(self,name,sep,axis,pp):
    """Check to see if the axis makes sense"""
    n,o,d,label,unit=sep.axis(axis)
    if n>1: 
      n2=self.pars.par("a%s_n"%pp)[0]
      o2=self.pars.par("a%s_o"%pp)[0]
      d2=self.pars.par("a%s_d"%pp)[0]
      SEP.sepfile.axis_same("%s(%s) axis %d"%(name,sep.name,axis),n,o,d,pp,n2,o2,d2)

  def check_data(self,data):
    """Check to see if the data makes sense"""
    if self.data_check.has_key(data.name): return
    if data.headers or data.grid:
      SEP.util.err("Data %s should be a regular dataset"%data.name)
    self.axis_check("Data",data,1,"mx")
    self.axis_check("Data",data,2,"my")
    self.axis_check("Data",data,3,"hx")
    self.axis_check("Data",data,4,"hy")
    self.axis_check("Data",data,5,"w_")
    self.data_check[data.name]=1
    


  def check_wavespace(self):
    """Check to make sure wavespace is described"""
    nmx,omx,dmx = self.check_wave_axis("mx")
    nmy,omy,dmy = self.check_wave_axis("my")
    nhx,ohx,dhx = self.check_wave_axis("hx")
    nhy,ohy,dhy = self.check_wave_axis("hy")
    nz,oz,dz = self.check_wave_axis("z_")
    nw,ow,dw = self.check_wave_axis("w_")

    bm,em= SEP.sepfile.min_max(nmx,omx,dmx)
    bh,eh= SEP.sepfile.min_max(nhx,ohx,dhx)
    if abs(bh) > abs(eh) : mx=abs(bh)
    else: mx=abs(eh)
    self.pars.add_param("wave_x_min",bm-mx)
    self.pars.add_param("wave_x_max",em+mx)

    bm,em= SEP.sepfile.min_max(nmy,omy,dmy)
    bh,eh= SEP.sepfile.min_max(nhy,ohy,dhy)
    if abs(bh) > abs(eh) : mx=abs(bh)
    else: mx=abs(eh)
    self.pars.add_param("wave_y_min",bm-mx)
    self.pars.add_param("wave_y_max",em+mx)

    z,err=self.pars.par("nzb",1)
    z=int(z)
    if nz/z != int(float(nz)/float(z)): 
      SEP.util.err("nzb(%d) is not a factor of %d "%(z,nz))
    self.pars.add_param("nw_tot",nw)

  def check_wave_axis(self,par,ndef=None,odef=None,ddf=None):
    """Return the n,o,d [with possible defaults for a wavespace axis"""
    n,err=self.pars.par("a%s_n"%par,ndef)
    if err==0: SEP.util.err("Required parameter a%s_n missing"%par)
    o,err=self.pars.par("a%s_o"%par,ndef)
    if err==0: SEP.util.err("Required parameter a%s_o missing"%par)
    d,err=self.pars.par("a%s_d"%par,ndef)
    if err==0: SEP.util.err("Required parameter a%s_d missing"%par)
    return int(n),float(o),float(d)

  def create_model(self,name):
    """Create the model structure"""
    do_real=self.pars.par("image_real","n")[0]
    if str(do_real)[0]=="y" or  str(do_real)[0]=="1":do_real=1
    else: do_real=None
    return self.create_file(name,["mx","my","hx","hy","z_"],do_real)

  def create_data(self,name):
    """Create the data structure"""
    return self.create_file(name,["mx","my","hx","hy","w_"])

  def create_file(self,name,axes,do_real=None):
    if do_real: junk=SEP.sep_vector.sfloat(name=name,space_only=1)
    else:       junk=SEP.sep_vector.scmplx(name=name,space_only=1)
    for i in range(len(axes)):
       n,o,d=self.get_wave_axis(axes[i])
       junk.set_axis(i+1,n,o,d,axes[i],"none")
    if do_real:junk.history.add_param("esize",4)
    else:junk.history.add_param("esize",8)
    return junk

  def get_wave_axis(self,par):
    """Return aX_n, aX_o, aX_d"""
    n=self.pars.par("a%s_n"%par)[0] 
    o=self.pars.par("a%s_o"%par)[0] 
    d=self.pars.par("a%s_d"%par)[0] 
    return n,o,d
  
       

class wei(wei_operator):
  """Basic wei object"""
  def __init__(self,pars,prog=None,code=None,**kw):
    self.code=code
    temp_dir= os.getcwd()+os.sep+"tmp"
    if not os.path.isfile:
      self.code.build(temp_dir)
    SEP.spawn.run_wait("rm -rf temp_dir".split())
    if not prog: prog=self.code.na,e
    wei_operator.__init__(self,prog,pars,**kw)

class wei_ang(wei):
  """Class to do wem with ph gathers"""
  def __init__(self,pars,prog=None, code=None,**kw):
    self.code=code
    temp_dir= os.getcwd()+os.sep+"tmp"
    if not os.path.isfile:
      self.code.build(temp_dir)
    SEP.spawn.run_wait("rm -rf temp_dir".split())
    if not prog: prog=self.code.build(temp_dir)
    pars.add_string("ang_par=y")
    wei_operator.__init__(self,prog,pars,**kw)

  def check_model(self,model):
    """Check to see if the model makes sense"""
    if self.model_check.has_key(model.name): return
    if model.headers or model.grid:
      SEP.util.err("Model %s should be a regular dataset"%model.name)
    self.axis_check("Model",model,1,"mx")
    self.axis_check("Model",model,2,"my")
    n,o,d,label,unit=model.axis(3)
    n2=int(self.pars.par("apx_n",error=1)[0])
    n3=int(self.pars.par("ahx_n",error=1)[0])
    if n!=n2 or n!=n3: 
      SEP.util.err("Length of  axis 3 (%d) of model, apx_n(%d),ahx_n(%d) must be the same"%(n,n2,n3))
    n,o,d,label,unit=model.axis(4)
    n2=int(self.pars.par("apy_n",error=1)[0])
    n3=int(self.pars.par("ahy_n",error=1)[0])
    if n!=n2 or n!=n3: 
      SEP.util.err("Length of  axis 4 of model, apy_n,apy_n must be the same")
    self.axis_check("Model",model,5,"z_")
    self.model_check[model.name]=1

  def create_model(self,name):
    """Create the model structure"""
    do_real=self.pars.par("image_real","n")[0]
    if str(do_real)[0]=="y" or  str(do_real)[0]=="1":do_real=1
    else: do_real=None
    return self.create_file(name,["mx","my","px","py","z_"],do_real)


class wei_migration(wei):
  """WEM  migration"""
  def __init__(self,pars,prog=None,code=None,**kw):
    pars.add_string("operation=migration")
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    wei.__init__(self,pars,prog,code,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    if err1==0: SEP.util.err("When doing migration must provide model (R)")
    if err2==0: SEP.util.err("When doing migration must provide data (D)")
    self.data=SEP.sep_vector.scmplx(tag=dtag)
    self.check_data(self.data)
    if self.verb: SEP.util.msg("Initializing/reading a model")
    self.image=self.create_model(mtag)
    if not restart: 
        self.image.write_file()
        self.image.zero()
    self.restart=restart
  def migrate(self):
    """Run a migration"""
    self.run_adjoint(self.image.name,self.data.name,restart=self.restart)

class wei_modeling(wei):
  """WEM  migration"""
  def __init__(self,pars,prog=None,code=None,**kw):
    pars.add_string("operation=modeling")
    wei.__init__(self,pars,prog,code,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    if err1==0: SEP.util.err("When doing modeling must provide model (R)")
    if err2==0: SEP.util.err("When doing modeling must provide data (D)")
    model=SEP.sepfile.sep_file(file=mtag)
    esize=model.history_par("esize",4)[0]
    if int(esize)==4: self.image=SEP.sep_vector.sfloat(tag=mtag)
    else: self.image=vector.scmplx(tag=mtag)
    self.image=SEP.sepfile.sep_file(file=mtag)
    self.check_model(model)
    if self.verb: SEP.util.msg("Initializing/reading a data")
    self.data=self.create_data(dtag)
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    if not restart: self.data.zero()
    self.restart=restart
  def model(self):
    """Run a modelling job"""
    self.run_forward(self.image.name,self.data.name,restart=self.restart)
  
class wei_ang_migration(wei_ang):
  """WEM  ph migration"""
  def __init__(self,pars,prog=None,code=None,**kw):
    pars.add_string("operation=migration")
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    wei_ang.__init__(self,pars,prog,code,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    if err1==0: SEP.util.err("When doing migration must provide model (R)")
    if err2==0: SEP.util.err("When doing migration must provide data (D)")
    self.data=SEP.sep_vector.scmplx(tag=dtag)
    self.check_data(self.data)
    if self.verb: SEP.util.msg("Initializing/reading a model")
    self.image=self.create_model(mtag)
    if not restart: 
        self.image.write_file()
        self.image.zero()
    self.restart=restart
  def migrate(self):
    """Run a migration"""
    self.adjoint_op(self.image,self.data,restart=self.restart)

class wei_ang_modeling(wei_ang):
  """WEM  ph modeling"""
  def __init__(self,pars,prog=None,code=None,**kw):
    pars.add_string("operation=modeling")
    if kw.has_key("restart"): restart=kw["restart"]
    else: restart=None
    wei_ang.__init__(self,pars,prog,code,**kw)
    mtag,err1=self.pars.par("R")
    dtag,err2=self.pars.par("D")
    print self.pars.return_par_string()
    if err1==0: SEP.util.err("When doing modeling must provide model (R)")
    if err2==0: SEP.util.err("When doing modeling must provide data (D)")
    self.model=SEP.sep_vector.scmplx(tag=mtag)
    self.check_model(self.model)
    if self.verb: SEP.util.msg("Initializing/reading a model")
    self.data=self.create_data(dtag)
    if not restart: 
        self.data.write_file()
        self.data.zero()
    self.restart=restart
  def modeling(self):
    """Run modeling"""
    self.forward_op(self.model,self.data,restart=self.restart)
    
